from pathlib import Path
from typing import Union, List
import numpy as np


def parse_specnum_output_file(output_file: Union[str, Path], analysis_range: List[float], range_type: str):
    """
    Parses the specnum_output.txt file.

    Parameters
    ----------
    output_file : Union[str, Path]
        Path to the specnum_output.txt file.
    analysis_range : List[float], optional
        A list of two elements `[start_percent, end_percent]` specifying the portion of the entire simulation
        to consider for analysis. The values should be between 0 and 100, representing percentages of the
        total simulated time or the total number of events, depending on `range_type`. For example,
        `[50, 100]` would analyze only the latter half of the simulation. Default is `[0.0, 100.0]`.
    range_type : str, optional
        Determines the dimension used when applying `analysis_range`:
        - `'time'`: The percentages in `analysis_range` refer to segments of the total simulated time.
        - `'nevents'`: The percentages in `analysis_range` refer to segments of the total number of simulated events.
        Default is `'time'`.
    Returns
    -------
    Tuple[np.ndarray, List[str]]
        A tuple containing:
        - data_slice: The sliced data as a NumPy array.
        - header: List of header column names.
    """
    output_file = Path(output_file)
    if not output_file.is_file():
        raise FileNotFoundError(f"Output file '{output_file}' does not exist.")

    if range_type == 'time':
        column_index = 2
    elif range_type == 'nevents':
        column_index = 1
    else:
        raise ValueError("'range_type' must be either 'time' or 'nevents'")

    if not (isinstance(analysis_range, list) and len(analysis_range) == 2):
        raise ValueError("'analysis_range' must be a list with two elements.")
    if not all(isinstance(x, (int, float)) for x in analysis_range):
        raise ValueError("Both elements in 'analysis_range' must be numbers (int or float).")
    if not (0 <= analysis_range[0] <= 100 and 0 <= analysis_range[1] <= 100):
        raise ValueError("Values in 'analysis_range' must be between 0 and 100.")
    if analysis_range[0] > analysis_range[1]:
        raise ValueError("The first value in 'analysis_range' must be less than or equal to the second value.")

    # Read the header
    with output_file.open('r') as infile:
        header_line = infile.readline()
        header = header_line.strip().split()

    # Read the data
    try:
        data = np.loadtxt(str(output_file), skiprows=1)
    except Exception as e:
        raise ValueError(f"Could not read data from '{output_file}': {e}")

    if data.size == 0:
        raise ValueError(f"No data found in '{output_file}'.")

    # Ensure data is at least 2D (in case of single row)
    if data.ndim == 1:
        data = data[np.newaxis, :]

    column = data[:, column_index]
    final_value = column[-1]

    value_initial_percent = analysis_range[0] / 100.0 * final_value
    value_final_percent = analysis_range[1] / 100.0 * final_value

    data_slice = data[(column >= value_initial_percent) & (column <= value_final_percent)]

    return data_slice, header



import warnings
import numpy as np
from zacrostools.kmc_output import KMCOutput


def detect_issues(job_path,
                  analysis_range,
                  range_type='time',
                  energy_slope_thr=5.0e-10,
                  time_r2_thr=0.95,
                  max_points=100):
    """
    Detect potential issues in a KMC simulation by analyzing the lattice energy slope
    and the linearity of time with respect to the number of events. If has_issues = True,
    the simulation probably needs to run for longer in order to reach steady-state.

    Parameters
    ----------
    job_path : str
        Path to the directory containing KMC simulation output.
    analysis_range : List[float], optional
        A list of two elements `[start_percent, end_percent]` specifying the portion of the entire simulation
        to consider for analysis. The values should be between 0 and 100, representing percentages of the
        total simulated time or the total number of events, depending on `range_type`. For example,
        `[50, 100]` would analyze only the latter half of the simulation. Default is `[0.0, 100.0]`.
    range_type : str, optional
        Determines the dimension used when applying `analysis_range`:
        - `'time'`: The percentages in `analysis_range` refer to segments of the total simulated time.
        - `'nevents'`: The percentages in `analysis_range` refer to segments of the total number of simulated events.
        Default is `'time'`.
    energy_slope_thr : float, optional
        Threshold for the absolute energy slope (in eV/Å²/step) above
        which the simulation is considered to have an issue, by default 5.0e-10.
    time_r2_thr : float, optional
        Threshold for the R² value in the time vs. KMC events regression,
        by default 0.95. If the R² is below this value, the simulation is
        considered to have an issue.
    max_points : int, optional
        Maximum number of data points to use in the regression. If the data
        arrays are longer, they will be uniformly sampled down to this length,
        by default 100.
    """

    def reduce_size(time, energy, nevents, size):
        """
        Uniformly sample the data arrays down to a specified size.
        """
        length = len(nevents)
        if length <= size:
            return time, energy, nevents

        indices = np.round(np.linspace(0, length - 1, size)).astype(int)
        return time[indices], energy[indices], nevents[indices]

    # Get simulation data
    kmc_output = KMCOutput(job_path=job_path, analysis_range=analysis_range, range_type=range_type)

    # Reduce arrays if necessary
    time_reduced, energy_reduced, nevents_reduced = reduce_size(kmc_output.time,
                                                               kmc_output.energy,
                                                               kmc_output.nevents,
                                                               max_points)

    # Edge-case check: Need at least 2 data points to perform any linear regression
    if len(nevents_reduced) < 2:
        warnings.warn(f"Not enough data points to perform regression for {job_path}. Got {len(nevents_reduced)} points. Marking simulation as having issues.", UserWarning)
        return True

    # Check for trend in energy using linear regression
    coeffs_energy = np.polyfit(nevents_reduced, energy_reduced, 1)
    slope_energy = coeffs_energy[0]
    energy_trend = abs(slope_energy) > energy_slope_thr

    # Perform linear regression on time vs. nevents
    coeffs_time = np.polyfit(nevents_reduced, time_reduced, 1)
    slope_time = coeffs_time[0]
    intercept_time = coeffs_time[1]
    time_predicted = slope_time * nevents_reduced + intercept_time

    # Compute R² for the time vs. nevents regression
    r_squared_time = np.corrcoef(time_reduced, time_predicted)[0, 1] ** 2
    time_not_linear = r_squared_time < time_r2_thr

    # If either the energy slope is too large or time is not linear, flag issues
    has_issues = energy_trend or time_not_linear

    return has_issues



from zacrostools.kmc_model import KMCModel
from zacrostools.gas_model import GasModel
from zacrostools.energetics_model import EnergeticsModel
from zacrostools.reaction_model import ReactionModel
from zacrostools.lattice_model import LatticeModel

gas_model = GasModel.from_csv(csv_path='csv_files_for_tests/gas_model.csv')
energetics_model = EnergeticsModel.from_csv(csv_path='csv_files_for_tests/energetics_model.csv')
reaction_model = ReactionModel.from_csv(csv_path='csv_files_for_tests/reaction_model.csv')
lattice_model = LatticeModel(lattice_type='periodic_cell',
                             cell_vectors=((3.27, 0), (0, 3.27)),
                             sites={'tC': (0.25, 0.25), 'tM': (0.75, 0.75)},
                             copies=[10, 10],
                             neighboring_structure='from_distances',
                             max_distances={'tC-tC': 4.0, 'tC-tM': 4.0, 'tM-tM': 4.0, 'Pt-Pt': 4.0, 'Pt-tC': 4.0,
                                            'Pt-tM': 4.0},
                             )
lattice_model.repeat_lattice_model(4, 4)
for coordinates in [(0.3125, 0.3125), (0.3125, 0.5625), (0.5625, 0.3125), (0.5625, 0.5625)]:
    lattice_model.change_site_type(direct_coords=coordinates, new_site_type='Pt')
lattice_model.remove_site(direct_coords=(0.4375, 0.4375))
lattice_model.copies = [2, 2]

kmc_model = KMCModel(
    gas_model=gas_model,
    reaction_model=reaction_model,
    energetics_model=energetics_model,
    lattice_model=lattice_model)

stiffness_scalable_steps = [
    'aCH4_HfC', 'aCH4_Pt-1', 'aCH4_Pt-2', 'aCH4_in-1', 'aCH4_in-2',
    'aCO2_HfC', 'aCO2_Pt', 'bCH3_HfC', 'bCH3_Pt-1', 'bCH3_Pt-2',
    'bCH3_in-1', 'bCH3_in-2', 'bCH2_HfC', 'bCH2_in-1', 'bCH2_in-2',
    'bCH_HfC', 'bCH_in-1', 'bCH_in-2', 'bCH_in-3', 'bCH_in-4',
    'fCO_HfC', 'fCO_in-1', 'fCO_in-2', 'fCO_in-3', 'fCO_in-4',
    'CHtoCHO_HfC', 'CHtoCHO_in-1', 'CHtoCHO_in-2', 'CHtoCHO_in-3',
    'CHtoCHO_in-4', 'CHOtoCO_HfC', 'CHOtoCO_Pt-1', 'CHOtoCO_Pt-2',
    'CHOtoCO_in-1', 'CHOtoCO_in-2', 'CtoCOH_HfC', 'CtoCOH_in-1',
    'CtoCOH_in-2', 'COHtoCHO_HfC', 'COHtoCHO_Pt', 'bCO2_HfC',
    'bCO2_Pt', 'bCO2_in', 'fOH_HfC', 'fOH_Pt-1', 'fOH_Pt-2',
    'fOH_in_a-1', 'fOH_in_a-2', 'fOH_in_b-1', 'fH2O_HfC',
    'fH2O_Pt-1', 'fH2O_Pt-2', 'fH2O_in-1', 'fH2O_in-2',
    'CO2toCOOH_HfC', 'CO2toCOOH_Pt', 'CO2toCOOH_in-1',
    'CO2toCOOH_in-2', 'COOHtoCO_HfC', 'COOHtoCO_Pt-1',
    'COOHtoCO_Pt-2', 'COOHtoCO_in-1', 'COOHtoCO_in-2',
    'dO_in-1', 'dO_in-2', 'dH_in', 'dCO_in', 'dOH_in',
    'dH2O_in', 'dCH2_in-1', 'dCH2_in-2', 'dCH_in-1',
    'dCH_in-2', 'dCH_in-3', 'dCH_in-4', 'dCHO_in-1',
    'dCHO_in-2', 'dCOH_in-1', 'dCOH_in-2'
]

stiffness_scalable_symmetric_steps = [
    'dO_HfC', 'dH_HfC', 'dCO_HfC', 'dOH_HfC', 'dH2O_HfC',
    'dCH3_HfC', 'dCH2_HfC', 'dCH_HfC', 'dC_HfC', 'dCHO_HfC',
    'dCOH_HfC', 'dO_Pt', 'dH_Pt', 'dCO_Pt', 'dOH_Pt',
    'dH2O_Pt', 'dCH3_Pt', 'dCH2_Pt', 'dCHO_Pt', 'dCOH_Pt'
]

pressure = {'CH4': 0.0001, 'CO2': 1.0, 'CO': 0.0, 'H2': 0.0, 'H2O': 0.0, 'O2': 0.0}

reporting_scheme = {'snapshots': 'on event 100000',
                    'process_statistics': 'on event 100000',
                    'species_numbers': 'on event 100000'}

stopping_criteria = {'max_steps': 'infinity', 'max_time': 5.0e+06, 'wall_time': 172000}

auto_scaling_tags = {'check_every': 500000,
                     'min_separation': 50.0,
                     'max_separation': 100.0,
                     'tol_part_equil_ratio': 0.05,
                     'upscaling_factor': 5.0,
                     'upscaling_limit': 100.0,
                     'downscaling_limit': 2.0,
                     'min_noccur': 10}

kmc_model.create_job_dir(job_path='reference_files',
                         temperature=1000,
                         pressure=pressure,
                         reporting_scheme=reporting_scheme,
                         stopping_criteria=stopping_criteria,
                         stiffness_scaling_algorithm='prats2024',
                         stiffness_scalable_steps=stiffness_scalable_steps,
                         stiffness_scalable_symmetric_steps=stiffness_scalable_symmetric_steps,
                         stiffness_scaling_tags=auto_scaling_tags,
                         sig_figs_energies=3,
                         sig_figs_pe=3)



import os
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, Optional
import matplotlib.patheffects as pe
from matplotlib.colors import Normalize

from zacrostools.kmc_output import KMCOutput
from zacrostools.custom_exceptions import PlotError
from zacrostools.heatmaps.heatmap_functions import get_axis_label, extract_value


def plot_cputime(
        # general mandatory parameters
        ax,
        x: str,
        y: str,
        scan_path: str,
        # plot-specific optional parameters
        levels: Optional[Union[list, np.ndarray]] = None,
        # general optional parameters
        cmap: str = "inferno",
        show_points: bool = False,
        show_colorbar: bool = True,
        auto_title: bool = False):
    """
    Plots a CPU time heatmap using a contourf plot with linear scaling.

    The CPU time (elapsed CPU time) is read from each simulation’s KMCOutput.
    Linear normalization is applied.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axis object where the plot is drawn.
    x : str
        Parameter for the x-axis (e.g., 'pressure_X' or 'temperature').
    y : str
        Parameter for the y-axis.
    scan_path : str
        Path to the directory containing simulation subdirectories.
    levels : list or np.ndarray, optional
        Contour levels for the plot. If provided, normalization uses
        vmin=min(levels) and vmax=max(levels). Default is None.
    cmap : str, optional
        Colormap to be used for the plot (default is 'inferno').
    show_points : bool, optional
        If True, overlay the grid points on the heatmap. Default is False.
    show_colorbar : bool, optional
        If True, display a colorbar alongside the heatmap. Default is True.
    auto_title : bool, optional
        If True, automatically set a title for the plot. Default is False.

    Notes
    -----
    - CPU time values are extracted from each simulation's KMCOutput; if a simulation
      cannot be read, NaN is assigned.
    - A linear normalization is used to map CPU time values onto the color scale.
    """

    # Validate scan_path
    if not os.path.isdir(scan_path):
        raise ValueError(f"Scan path folder does not exist: {scan_path}")

    simulation_dirs = glob.glob(os.path.join(scan_path, "*"))
    if len(simulation_dirs) == 0:
        raise ValueError(f"Scan path folder is empty: {scan_path}")

    # Determine whether x and y values are logarithmic (based on presence of 'pressure' in the variable name)
    x_is_log = "pressure" in x
    y_is_log = "pressure" in y

    # Initialize lists and DataFrame to store data
    x_value_list, y_value_list = [], []
    df = pd.DataFrame()

    # Collect simulation data in a dataframe
    for sim_path in simulation_dirs:
        folder_name = os.path.basename(sim_path)

        # Extract x and y values
        x_value = extract_value(x, sim_path)
        y_value = extract_value(y, sim_path)
        df.loc[folder_name, "x_value"] = x_value
        df.loc[folder_name, "y_value"] = y_value
        if x_value not in x_value_list:
            x_value_list.append(x_value)
        if y_value not in y_value_list:
            y_value_list.append(y_value)

        # Initialize KMCOutput and retrieve CPU time
        try:
            kmc_output = KMCOutput(path=sim_path)
            df.loc[folder_name, 'cpu_time'] = kmc_output.cpu_time
        except Exception as e:
            print(f"Warning: Could not initialize KMCOutput for {folder_name}: {e}")
            df.loc[folder_name, "cpu_time"] = float('NaN')

    # Build sorted arrays for x and y axis values
    x_value_list = np.sort(np.asarray(x_value_list))
    y_value_list = np.sort(np.asarray(y_value_list))
    x_list = np.power(10, x_value_list) if x_is_log else x_value_list
    y_list = np.power(10, y_value_list) if y_is_log else y_value_list

    # Create a 2D grid for CPU time values
    z_axis = np.full((len(y_value_list), len(x_value_list)), np.nan)
    for i, x_val in enumerate(x_value_list):
        for j, y_val in enumerate(y_value_list):
            matching_indices = df[(df['x_value'] == x_val) & (df['y_value'] == y_val)].index
            if len(matching_indices) > 1:
                raise PlotError(
                    f"Several folders have the same values of {x} ({x_val}) and {y} ({y_val})")
            elif len(matching_indices) == 0:
                print(f"Warning: folder for x = {x_val} and y = {y_val} missing, NaN assigned")
            else:
                folder_name = matching_indices[0]
                z_axis[j, i] = df.loc[folder_name, "cpu_time"]

    x_axis, y_axis = np.meshgrid(x_list, y_list)

    # Create a contour plot with linear normalization
    if levels is not None:
        cp = ax.contourf(x_axis, y_axis, z_axis, levels=levels, cmap=cmap,
                         norm=Normalize(vmin=min(levels), vmax=max(levels)))
    else:
        cp = ax.contourf(x_axis, y_axis, z_axis, cmap=cmap)

    if show_colorbar:
        cbar = plt.colorbar(cp, ax=ax)

    ax.set_xlim(np.min(x_list), np.max(x_list))
    ax.set_ylim(np.min(y_list), np.max(y_list))
    ax.set_xscale('log' if x_is_log else 'linear')
    ax.set_yscale('log' if y_is_log else 'linear')
    ax.set_xlabel(get_axis_label(x))
    ax.set_ylabel(get_axis_label(y))
    ax.set_facecolor("lightgray")

    if auto_title:
        ax.set_title(
            label="CPU time ($s$)",
            y=1.0,
            pad=-14,
            color="w",
            path_effects=[pe.withStroke(linewidth=2, foreground="black")]
        )

    if show_points:
        ax.plot(x_axis.flatten(), y_axis.flatten(), 'w.', markersize=3)

    return cp



import pytest
import shutil
import os


@pytest.fixture
def job_dir(tmp_path):
    """
    Fixture to provide a job directory path within a temporary path.
    Does NOT create the directory, allowing create_job_dir to handle it.
    """
    jd = tmp_path / "job"
    # Do NOT create the directory here
    return jd


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    Hook to execute after each test to check for failures and copy the job directory if the test fails.
    """
    # Execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # Only act on actual test function calls, not setup/teardown
    if rep.when == "call" and rep.failed:
        # Access the 'job_dir' from the test function's arguments
        job_dir = item.funcargs.get('job_dir')
        if job_dir and job_dir.exists():
            # Define a path to copy the failed job_dir for inspection
            failed_tests_root = os.path.join(os.path.dirname(__file__), 'failed_tests')
            os.makedirs(failed_tests_root, exist_ok=True)
            failed_dir = os.path.join(failed_tests_root, item.name)

            # Remove the failed_dir if it already exists to avoid shutil.copytree error
            if os.path.exists(failed_dir):
                shutil.rmtree(failed_dir)

            shutil.copytree(job_dir, failed_dir)
            print(f"\nTest failed. Job directory copied to: {failed_dir}")



import numpy as np
import pandas as pd
from pathlib import Path
from typing import Union, Dict, Any, List, Optional


def parse_general_output_file(
        output_file: Union[str, Path],
        parse_stiffness_coefficients: bool = False
) -> Dict[str, Any]:
    """
    Parses the general_output.txt file from a Zacros simulation and extracts
    various simulation parameters.

    Parameters
    ----------
    output_file : Union[str, Path]
        Path to the general_output.txt file.
    parse_stiffness_coefficients : bool, optional
        If True, parse stiffness coefficients from the Commencing simulation block.
        Default is False.

    Returns
    -------
    data : Dict[str, Any]
        A dictionary containing parsed information including:
            'version': str
            From Simulation setup:
                'random_seed': int
                'temperature': float
                'pressure': float
                'n_gas_species': int
                'gas_specs_names': List[str]
                'gas_energies': List[float]
                'gas_molar_fracs': List[float]
                'n_surf_species': int
                'surf_specs_names': List[str]
                'surf_specs_dent': List[int]
                'enable_stiffness_scaling': bool
                'stiffness_scaling_algorithm': Optional[str]  # 'prats2024', 'legacy', or None
                'stiffness_scaling_tags': Optional[Dict[str, Union[int, float]]]
            From Lattice setup:
                'area': float
                'n_sites': int
                'n_site_types': int
                'site_types': Dict[str, int]
            From Mechanism setup:
                'n_steps': int
                # If enable_stiffness_scaling=True and algorithm='legacy':
                #   'stiffness_scalable_steps': List[str]
                # If enable_stiffness_scaling=True and algorithm='prats2024':
                #   'non_symmetric_stiffness_scalable_steps': List[str]
                #   'symmetric_stiffness_scalable_steps': List[str]
                #   'stiffness_scalable_steps': List[str]
    """


    output_file = Path(output_file)
    with output_file.open('r') as f:
        lines = f.readlines()

    data: Dict[str, Any] = {}

    # Define known blocks for reference
    known_blocks = [
        "Compiler information:",
        "Threading/multiprocessing information:",
        "Simulation setup:",
        "Lattice setup:",
        "Energetics setup:",
        "Mechanism setup:",
        "Initial state setup:",
        "Preparing simulation:",
        "Commencing simulation:",
        "Simulation stopped:",
        "Performance facts:",
        "Execution queue statistics:",
        "Memory usage statistics:"
    ]

    # Helper function to find block lines
    def find_block(full_lines: List[str], block_name: str, blocks: List[str]):
        start = None
        end = None
        for i, line in enumerate(full_lines):
            if line.strip() == block_name:
                start = i
            elif start is not None and line.strip() in blocks and line.strip() != block_name:
                end = i
                break
        if start is not None and end is None:
            end = len(full_lines)
        return start, end

    # Parsing functions
    def parse_header(full_lines: List[str]) -> str:
        """
        Parse the header of the file to extract the version.
        """
        version = None
        for line in full_lines:
            if "ZACROS" in line:
                parts = line.strip("|").strip().split()
                if "ZACROS" in parts:
                    idx = parts.index("ZACROS")
                    if idx + 1 < len(parts):
                        version = parts[idx + 1]
                        break
        return version

    def parse_compiler_information(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Compiler information:", blocks)
        # For now, do nothing

    def parse_threading_information(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Threading/multiprocessing information:", blocks)
        # For now, do nothing

    def parse_simulation_setup(full_lines: List[str], blocks: List[str]) -> Dict[str, Any]:
        start, end = find_block(full_lines, "Simulation setup:", blocks)

        data_block: Dict[str, Any] = {
            'random_seed': None,
            'temperature': None,
            'pressure': None,
            'n_gas_species': None,
            'gas_specs_names': None,
            'gas_energies': None,
            'gas_molar_fracs': None,
            'n_surf_species': None,
            'surf_specs_names': None,
            'surf_specs_dent': None,
            'enable_stiffness_scaling': False,
            'stiffness_scaling_algorithm': None,
            'stiffness_scaling_tags': None  # Initialize as None
        }

        if start is not None and end is not None:
            for line in full_lines[start:end]:
                line_stripped = line.strip()
                if not line_stripped:
                    continue

                if line_stripped.startswith("Random sequence with seed:"):
                    parts = line_stripped.split(":")
                    if len(parts) == 2:
                        data_block['random_seed'] = int(parts[1].strip())

                elif line_stripped.startswith("Temperature:"):
                    parts = line_stripped.split(":")
                    if len(parts) == 2:
                        data_block['temperature'] = float(parts[1].strip())

                elif line_stripped.startswith("Pressure:"):
                    parts = line_stripped.split(":")
                    if len(parts) == 2:
                        data_block['pressure'] = float(parts[1].strip())

                elif line_stripped.startswith("Number of gas species:"):
                    parts = line_stripped.split(":")
                    if len(parts) == 2:
                        data_block['n_gas_species'] = int(parts[1].strip())

                elif line_stripped.startswith("Gas species names:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['gas_specs_names'] = parts[1].split()

                elif line_stripped.startswith("Gas species energies:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['gas_energies'] = [float(x) for x in parts[1].split()]

                elif line_stripped.startswith("Gas species molar fractions:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['gas_molar_fracs'] = [float(x) for x in parts[1].split()]

                elif line_stripped.startswith("Number of surface species:"):
                    parts = line_stripped.split(":")
                    if len(parts) == 2:
                        data_block['n_surf_species'] = int(parts[1].strip())

                elif line_stripped.startswith("Surface species names:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        surf_names = parts[1].split()
                        data_block['surf_specs_names'] = surf_names

                elif line_stripped.startswith("Surface species dentation:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['surf_specs_dent'] = [int(x) for x in parts[1].split()]

                elif line_stripped.startswith("Keyword enable_stiffness_scaling (prats2024) parsed."):
                    data_block['enable_stiffness_scaling'] = True
                    data_block['stiffness_scaling_algorithm'] = 'prats2024'

                elif line_stripped.startswith("Keyword enable_stiffness_scaling (legacy) parsed."):
                    data_block['enable_stiffness_scaling'] = True
                    data_block['stiffness_scaling_algorithm'] = 'legacy'

                # Parse Stiffness Scaling Tags
                if data_block['enable_stiffness_scaling']:
                    # Initialize the tags dictionary
                    if data_block['stiffness_scaling_tags'] is None:
                        data_block['stiffness_scaling_tags'] = {}

                    # Define mapping based on the algorithm
                    if data_block['stiffness_scaling_algorithm'] == 'prats2024':
                        if line_stripped.startswith("[Stiffness scaling] Frequency of stiffness checks:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['check_every'] = int(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Minimum allowed separation of time-scales between fastest non-quasi-equilibrated step and slowest quasi-equilibrated one:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['min_separation'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Maximum allowed separation of time-scales between fastest non-quasi-equilibrated step and slowest quasi-equilibrated one:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['max_separation'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Tolerance on partial-equilibrium ratio for detecting quasi-equilibrated steps:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['tol_part_equil_ratio'] = float(value)

                        elif line_stripped.startswith("[Stiffness scaling] Upscaling_factor is set as:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['upscaling_factor'] = float(value)

                        elif line_stripped.startswith("[Stiffness scaling] Upscaling_limit is set as:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['upscaling_limit'] = float(value)

                        elif line_stripped.startswith("[Stiffness scaling] Downscaling_limit is set as:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['downscaling_limit'] = float(value)

                        elif line_stripped.startswith("[Stiffness scaling] min_noccur enevnts:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['min_noccur'] = int(value)

                    elif data_block['stiffness_scaling_algorithm'] == 'legacy':
                        if line_stripped.startswith("[Stiffness scaling] Frequency of stiffness checks:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['check_every'] = int(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Minimum allowed separation of time-scales between fastest non-quasi-equilibrated step and slowest quasi-equilibrated one:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['min_separation'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Maximum allowed separation of time-scales between fastest non-quasi-equilibrated step and slowest quasi-equilibrated one:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['max_separation'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Tolerance on partial-equilibrium ratio for detecting quasi-equilibrated steps:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['tol_part_equil_ratio'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Threshold on stiffness coefficient for applying scaling:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['stiffn_coeff_threshold'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Factor for scaling up/down the rate constants:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['scaling_factor'] = float(value)

                        elif line_stripped.startswith(
                                "[Stiffness scaling] Default maximum allowed separation of time-scales between fastest and slowest quasi-equilibrated steps:"):
                            value = line_stripped.split(":", 1)[1].strip()
                            data_block['stiffness_scaling_tags']['max_qequil_separation'] = float(value)

        return data_block

    def parse_lattice_setup(full_lines: List[str], blocks: List[str], version_str: str) -> Dict[str, Any]:
        """
        Parse the "Lattice setup:" block.

        We need to extract:
            'area' (float)
            'n_sites' (int)
            'n_site_types' (int)
            'site_types' (dict: name->count)
        """
        start, end = find_block(full_lines, "Lattice setup:", blocks)
        data_block: Dict[str, Any] = {
            'area': None,
            'n_sites': None,
            'n_site_types': None,
            'site_types': {}
        }

        if version_str is not None:
            version_val = float(version_str)
        else:
            version_val = 5.0  # default assumption if not found

        # Identify which lines to look for based on version
        if version_val < 3.0:
            area_key = "Surface area:"
            n_sites_key = "Number of lattice sites:"
            site_type_header = "Site type names and number of sites of that type:"
        else:
            area_key = "Lattice surface area:"
            n_sites_key = "Total number of lattice sites:"
            site_type_header = "Site type names and total number of sites of that type:"

        if start is not None and end is not None:
            read_site_types = False
            for line in full_lines[start:end]:
                line_stripped = line.strip()
                if not line_stripped:
                    continue

                if line_stripped.startswith(area_key):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['area'] = float(parts[1].strip())

                elif line_stripped.startswith(n_sites_key):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['n_sites'] = int(parts[1].strip())

                elif line_stripped.startswith("Number of site types:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['n_site_types'] = int(parts[1].strip())

                elif line_stripped.startswith(site_type_header):
                    read_site_types = True
                    continue

                elif read_site_types:
                    if line_stripped.startswith("Finished reading lattice input.") or \
                            line_stripped.startswith("Maximum coordination number:"):
                        read_site_types = False
                    else:
                        # Parse lines like: tC (100)
                        if '(' in line_stripped and ')' in line_stripped:
                            name_part, num_part = line_stripped.split('(')
                            name_part = name_part.strip()
                            num_part = num_part.strip(' )')
                            data_block['site_types'][name_part] = int(num_part)

        return data_block

    def parse_energetics_setup(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Energetics setup:", blocks)
        # For now, ignore this block

    def parse_mechanism_setup(full_lines: List[str], blocks: List[str], enable_stiffness: bool,
                              algorithm: Optional[str]) -> Dict[str, Any]:
        """
        Parse the "Mechanism setup:" block.

        Extract:
            'n_steps': int
            If enable_stiffness_scaling == True:
                If algorithm == 'legacy':
                    'stiffness_scalable_steps': List[str]
                If algorithm == 'prats2024':
                    'non_symmetric_stiffness_scalable_steps': List[str]
                    'symmetric_stiffness_scalable_steps': List[str]
                    'stiffness_scalable_steps': List[str]
        """
        start, end = find_block(full_lines, "Mechanism setup:", blocks)

        data_block: Dict[str, Any] = {
            'n_steps': None
        }

        # Initialize variables for stiffness scaling
        data_block['stiffness_scalable_steps'] = None
        data_block['non_symmetric_stiffness_scalable_steps'] = None
        data_block['symmetric_stiffness_scalable_steps'] = None

        if start is not None and end is not None:
            read_legacy_stiff = False
            read_non_symmetric = False
            read_symmetric = False

            legacy_steps = []
            non_sym_steps = []
            sym_steps = []

            for line in full_lines[start:end]:
                line_stripped = line.strip()
                if not line_stripped:
                    continue

                if line_stripped.startswith("Number of elementary steps:"):
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        data_block['n_steps'] = int(parts[1].strip())

                if enable_stiffness and algorithm == 'legacy':
                    # Stiffness scaling enabled for the following elementary steps:
                    if line_stripped.startswith("Stiffness scaling enabled for the following elementary steps:"):
                        read_legacy_stiff = True
                        continue
                    elif read_legacy_stiff:
                        # Lines like: "Fwd/Rev: 1/2 - aO2_HfC_fwd/aO2_HfC_rev"
                        if line_stripped.startswith("Fwd/Rev:"):
                            # Extract the base step name
                            parts = line_stripped.split('-', 1)
                            if len(parts) == 2:
                                steps_part = parts[1].strip()
                                fwd_rev = steps_part.split('/')
                                base_name = fwd_rev[0].replace('_fwd', '').replace('_rev', '')
                                legacy_steps.append(base_name)
                        else:
                            # If we hit something else, we might have finished reading
                            # but let's rely on break conditions if needed.
                            pass

                if enable_stiffness and algorithm == 'prats2024':
                    # Stiffness scaling (non-symmetric) enabled...
                    if line_stripped.startswith(
                            "Stiffness scaling (non-symmetric) enabled for the following elementary steps:"):
                        read_non_symmetric = True
                        read_symmetric = False
                        continue
                    elif line_stripped.startswith(
                            "Stiffness scaling (symmetric) enabled for the following elementary steps:"):
                        read_non_symmetric = False
                        read_symmetric = True
                        continue
                    elif read_non_symmetric:
                        # read lines like Fwd/Rev: 1/2 - aO2_HfC_fwd/aO2_HfC_rev
                        if line_stripped.startswith("Fwd/Rev:"):
                            parts = line_stripped.split('-', 1)
                            if len(parts) == 2:
                                steps_part = parts[1].strip()
                                fwd_rev = steps_part.split('/')
                                base_name = fwd_rev[0].replace('_fwd', '').replace('_rev', '')
                                non_sym_steps.append(base_name)
                    elif read_symmetric:
                        if line_stripped.startswith("Fwd/Rev:"):
                            parts = line_stripped.split('-', 1)
                            if len(parts) == 2:
                                steps_part = parts[1].strip()
                                fwd_rev = steps_part.split('/')
                                base_name = fwd_rev[0].replace('_fwd', '').replace('_rev', '')
                                sym_steps.append(base_name)

            # After reading the block, store the lists if needed
            if enable_stiffness and algorithm == 'legacy':
                data_block['stiffness_scalable_steps'] = legacy_steps
            elif enable_stiffness and algorithm == 'prats2024':
                data_block['non_symmetric_stiffness_scalable_steps'] = non_sym_steps
                data_block['symmetric_stiffness_scalable_steps'] = sym_steps
                data_block['stiffness_scalable_steps'] = non_sym_steps + sym_steps

        return data_block

    def parse_preparing_simulation(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Preparing simulation:", blocks)
        # For now, do nothing

    def parse_commencing_simulation(
            full_lines: List[str],
            blocks: List[str],
            parse_stiffness_coefficients: bool,
            stiffness_scalable_steps: List[str],
            algorithm: Union[str, None]
    ) -> Dict[str, Any]:
        """
        Parses the "Commencing simulation:" block.

        Parameters
        ----------
        full_lines : List[str]
            All lines from the output file.
        blocks : List[str]
            List of known block headers to identify the end of the current block.
        parse_stiffness_coefficients : bool
            Flag to determine whether to parse stiffness scaling coefficients.
        stiffness_scalable_steps : List[str]
            List of step names that are stiffness scalable.
        algorithm : Union[str, None]
            The stiffness scaling algorithm used ('prats2024', 'legacy', or None).

        Returns
        -------
        Dict[str, Any]
            A dictionary with the key 'stiffness_scaling_coefficients' containing a Pandas DataFrame
            or None if parsing is not required.
        """

        if not parse_stiffness_coefficients:
            return {'stiffness_scaling_coefficients': None}

        start, end = find_block(full_lines, "Commencing simulation:", blocks)

        if start is None:
            # Block not found
            return {'stiffness_scaling_coefficients': None}

        # Extract the lines within the "Commencing simulation:" block
        block_lines = full_lines[start:end]

        data_dict = {
            'nevents': [0],
            'time': [0.0]
        }

        for step in stiffness_scalable_steps:
            data_dict[step] = [1.0]

        def parse_prats2024(block_lines: List[str], steps: List[str], data_dict: Dict[str, List[Any]]):
            """
            Parses sub-blocks specific to the 'prats2024' algorithm and updates data_dict.

            Parameters
            ----------
            block_lines : List[str]
                Lines within the "Commencing simulation:" block.
            steps : List[str]
                List of stiffness scalable steps.
            data_dict : Dict[str, List[Any]]
                Dictionary to store parsed data.
            """
            i = 0
            n = len(block_lines)

            # Skip the initial settings sub-block
            while i < n:
                line = block_lines[i].strip()
                if line.startswith("Stiffness scaling module invoked at time t = "):
                    break
                i += 1

            # Iterate through the remaining lines to parse each invocation sub-block
            while i < n:
                line = block_lines[i].strip()

                if line.startswith("Stiffness scaling module invoked at time t = "):
                    # Example line:
                    # "Stiffness scaling module invoked at time t = 1.2609349242516475E-006 and number of steps = 16000:"

                    # Extract the portion after the prefix
                    prefix = "Stiffness scaling module invoked at time t = "
                    rest = line[len(prefix):]

                    # Split to get time and number of steps
                    if " and number of steps = " in rest:
                        time_str, nevents_str = rest.split(" and number of steps = ")
                        try:
                            time = float(time_str)
                            nevents = int(nevents_str.rstrip(':'))
                        except ValueError:
                            # If conversion fails, skip this sub-block
                            i += 1
                            continue
                    else:
                        # If the expected format is not found, skip this sub-block
                        i += 1
                        continue

                    # Initialize coefficients to 1.0 for all steps
                    coefficients = {step: 1.0 for step in steps}

                    # Move to the next line after the header
                    i += 1

                    # Traverse lines until "Updated list of downscaled steps:" is found
                    while i < n:
                        current_line = block_lines[i].strip()
                        if current_line.startswith("Updated list of downscaled steps:"):
                            break
                        i += 1

                    if i >= n:
                        # If "Updated list of downscaled steps:" is not found, skip to next sub-block
                        break

                    # Move to the lines after "Updated list of downscaled steps:"
                    i += 1

                    if i >= n:
                        # If there are no lines after the header, skip
                        break

                    updated_line = block_lines[i].strip()

                    if updated_line == "All stiffness coefficients are one":
                        # No changes needed; coefficients remain 1.0
                        i += 1
                    else:
                        # Parse each line with "Elementary steps X/Y (step_name): value"
                        while i < n:
                            current_line = block_lines[i].strip()

                            if not current_line.startswith("Elementary steps"):
                                # End of the updated list
                                break

                            # Example line:
                            # "Elementary steps 3/4 (aCO_HfC): 1.0514441655341260E-002"

                            # Split the line at the colon to separate step info and value
                            if ":" in current_line:
                                step_info, value_str = current_line.split(":", 1)
                                step_info = step_info.strip()
                                value_str = value_str.strip()

                                # Extract step_name from step_info
                                # step_info format: "Elementary steps X/Y (step_name)"
                                start_idx = step_info.find('(')
                                end_idx = step_info.find(')')
                                if start_idx != -1 and end_idx != -1 and end_idx > start_idx:
                                    step_name = step_info[start_idx + 1:end_idx]
                                    try:
                                        value = float(value_str)
                                        if step_name in coefficients:
                                            coefficients[step_name] = value
                                    except ValueError:
                                        # If value conversion fails, ignore and continue
                                        pass

                            # Move to the next line
                            i += 1

                    # Append the parsed data to data_dict
                    data_dict['time'].append(time)
                    data_dict['nevents'].append(nevents)
                    for step in steps:
                        data_dict[step].append(coefficients.get(step, 1.0))

                else:
                    # If the line does not start a new sub-block, move to the next line
                    i += 1

        def parse_legacy(block_lines: List[str], steps: List[str], data_dict: Dict[str, List[Any]]):
            """
            Parses sub-blocks specific to the 'legacy' algorithm and updates data_dict.

            Parameters
            ----------
            block_lines : List[str]
                Lines within the "Commencing simulation:" block.
            steps : List[str]
                List of stiffness scalable steps.
            data_dict : Dict[str, List[Any]]
                Dictionary to store parsed data.
            """
            i = 0
            n = len(block_lines)

            # Initialize current coefficients for all steps to 1.0
            current_coefficients = {step: 1.0 for step in steps}

            # Define possible sub-block headers
            sub_block_headers = [
                "Stiffness detected at time t = ",
                "Stiffness scaling possibly too aggressive at time t = ",
                "Stiffness possible at time t = "
            ]

            while i < n:
                line = block_lines[i].strip()

                # Check if the line starts with any of the sub-block headers
                if any(line.startswith(header) for header in sub_block_headers):
                    # Extract the time from the header
                    # Find the index of 't = ' and extract the float value
                    try:
                        # Find the position of 't = '
                        t_index = line.find('t = ')
                        if t_index == -1:
                            raise ValueError("Time not found in sub-block header.")

                        # Extract the part after 't = '
                        time_part = line[t_index + len('t = '):]

                        # The time value ends at the first occurrence of ':' or space
                        # Find the first occurrence of ':' or space after 't = '
                        end_index = len(time_part)
                        for sep in [':', ' ']:
                            sep_idx = time_part.find(sep)
                            if sep_idx != -1:
                                end_index = min(end_index, sep_idx)
                        time_str = time_part[:end_index]
                        time = float(time_str)
                    except ValueError:
                        # If time extraction fails, skip this sub-block
                        i += 1
                        continue

                    # Initialize coefficients for this sub-block as a copy of current_coefficients
                    coefficients = current_coefficients.copy()

                    # Assign NaN to 'nevents' as it's not available in 'legacy' algorithm
                    nevents = np.nan

                    # Move to the next line to parse the sub-block
                    i += 1

                    while i < n:
                        current_line = block_lines[i].strip()

                        # Check if we've reached the start of a new sub-block
                        if any(current_line.startswith(header) for header in sub_block_headers):
                            break

                        # Check if the line starts with 'Elementary step'
                        if current_line.startswith("Elementary step"):
                            # Example line:
                            # "Elementary step 3 - aCO_HfC_fwd (2.0376387410451285E-003). Occurred 1808 times (too fast). StiffCoeffNew/Old = 0.57043908398242493"

                            # Split the line at the first '(' to extract step name and coefficient
                            try:
                                step_part, rest = current_line.split('(', 1)
                                coeff_part, _ = rest.split(')', 1)

                                # Extract step identifier
                                # step_part format: "Elementary step X - step_name"
                                step_info = step_part.strip()
                                # Extract step_name by splitting at ' - '
                                if ' - ' in step_info:
                                    _, step_full = step_info.split(' - ', 1)
                                    # step_full format: "aCO_HfC_fwd" or similar
                                    step_name_full = step_full.strip()
                                    # We only want the base step name (e.g., "aCO_HfC" from "aCO_HfC_fwd")
                                    if step_name_full.endswith('_fwd'):
                                        step_name = step_name_full[:-4]  # Remove '_fwd'
                                    elif step_name_full.endswith('_rev'):
                                        # Skip reverse directions
                                        i += 1
                                        continue
                                    else:
                                        # If not ending with '_fwd' or '_rev', skip
                                        i += 1
                                        continue

                                    # Parse the coefficient
                                    coefficient_str = coeff_part.strip()
                                    coefficient = float(coefficient_str)

                                    # Update the coefficient for the step
                                    if step_name in coefficients:
                                        coefficients[step_name] = coefficient
                            except (ValueError, IndexError):
                                # If any parsing error occurs, skip this line
                                pass

                        # Move to the next line
                        i += 1

                    # After parsing the sub-block, update the current_coefficients
                    current_coefficients = coefficients.copy()

                    # Append the parsed data to data_dict
                    data_dict['time'].append(time)
                    data_dict['nevents'].append(nevents)
                    for step in steps:
                        data_dict[step].append(current_coefficients.get(step, 1.0))

                else:
                    # If the line does not start a sub-block header, move to the next line
                    i += 1

        # Based on the algorithm, parse the relevant sub-blocks
        if algorithm == 'prats2024':
            parse_prats2024(block_lines, stiffness_scalable_steps, data_dict)
        elif algorithm == 'legacy':
            parse_legacy(block_lines, stiffness_scalable_steps, data_dict)
        else:
            raise ValueError(f"Unsupported algorithm '{algorithm}'. Expected 'prats2024' or 'legacy'.")

        stiffness_df = pd.DataFrame(data_dict)

        return {'stiffness_scaling_coefficients': stiffness_df}

    def parse_simulation_stopped(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Simulation stopped:", blocks)
        # For now, do nothing

    def parse_performance_facts(full_lines: List[str], blocks: List[str]) -> Dict[str, Any]:
        start, end = find_block(full_lines, "Performance facts:", blocks)
        perf_data: Dict[str, Any] = {'cpu_time': None}  # Otherwise KMCOutput will crash for unfinished simulations
        if start is not None and end is not None:
            for line in full_lines[start:end]:
                line_stripped = line.strip()
                if line_stripped.startswith("Elapsed CPU time:"):
                    # Expected format: "Elapsed CPU time:         60327.0391 seconds"
                    parts = line_stripped.split(":", 1)
                    if len(parts) == 2:
                        # Split again to remove the "seconds" part and convert the first token to float
                        time_str = parts[1].strip().split()[0]
                        try:
                            perf_data['cpu_time'] = float(time_str)
                        except ValueError:
                            perf_data['cpu_time'] = None
                    break
        return perf_data

    def parse_execution_queue_statistics(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Execution queue statistics:", blocks)
        # For now, do nothing

    def parse_memory_usage_statistics(full_lines: List[str], blocks: List[str]):
        start, end = find_block(full_lines, "Memory usage statistics:", blocks)
        # For now, do nothing

    # Parse header to get version
    data['version'] = parse_header(lines)

    # Parse blocks
    parse_compiler_information(lines, known_blocks)
    parse_threading_information(lines, known_blocks)
    sim_data = parse_simulation_setup(lines, known_blocks)
    data.update(sim_data)
    lattice_data = parse_lattice_setup(lines, known_blocks, data['version'])
    data.update(lattice_data)
    parse_energetics_setup(lines, known_blocks)
    mech_data = parse_mechanism_setup(lines, known_blocks, data['enable_stiffness_scaling'],
                                      data['stiffness_scaling_algorithm'])
    data.update(mech_data)
    parse_preparing_simulation(lines, known_blocks)
    comm_data = parse_commencing_simulation(
        full_lines=lines,
        blocks=known_blocks,
        parse_stiffness_coefficients=parse_stiffness_coefficients,
        stiffness_scalable_steps=data.get('stiffness_scalable_steps', []),
        algorithm=data.get('stiffness_scaling_algorithm', None)
    )
    data.update(comm_data)
    parse_simulation_stopped(lines, known_blocks)
    perf_data = parse_performance_facts(lines, known_blocks)
    data.update(perf_data)
    parse_execution_queue_statistics(lines, known_blocks)
    parse_memory_usage_statistics(lines, known_blocks)

    return data



import os
import numpy as np
from typing import Union
from zacrostools.simulation_input import parse_simulation_input_file
from zacrostools.general_output import parse_general_output_file
from zacrostools.specnum_output import parse_specnum_output_file
from zacrostools.custom_exceptions import enforce_types, KMCOutputError, EnergeticsModelError


class KMCOutput:

    """
    A class that represents a KMC (Kinetic Monte Carlo) simulation output.

    Attributes
    ----------
    area : float
        Lattice surface area (in Å²).
    av_coverage : Dict[str, float]
        Average coverage of surface species (in %). Example: `KMCOutput.av_coverage['CO']`.
    av_coverage_per_site_type : Dict[str, Dict[str, float]]
        Average coverage of surface species per site type (in %).
    av_energy : float
        Average lattice energy (in eV·Å⁻²).
    av_total_coverage : float
        Average total coverage of surface species (in %).
    av_total_coverage_per_site_type : Dict[str, float]
        Average total coverage of surface species per site type (in %).
    coverage : Dict[str, np.ndarray]
        Coverage of surface species over time (in %). Example: `KMCOutput.coverage['CO']`.
    coverage_per_site_type : Dict[str, Dict[str, np.ndarray]]
        Coverage of surface species per site type over time (in %).
    cpu_time : float
        Final elapsed cpu time (in seconds).
    dominant_ads : str
        Most dominant surface species, used for plotting kinetic phase diagrams.
    dominant_ads_per_site_type : Dict[str, str]
        Most dominant surface species per site type, used for plotting kinetic phase diagrams.
    energy : np.ndarray
        Lattice energy (in eV·Å⁻²).
    final_energy : float
        Final lattice energy (in eV·Å⁻²).
    finaltime : float
        Final simulated time (in seconds).
    gas_specs_names : List[str]
        Gas species names.
    n_gas_species : int
        Number of gas species.
    n_sites : int
        Total number of lattice sites.
    n_surf_species : int
        Number of surface species.
    nevents : np.ndarray
        Number of events occurred.
    production : Dict[str, np.ndarray]
        Gas species produced over time. Example: `KMCOutput.production['CO']`.
    surf_specs_names : List[str]
        Surface species names.
    time : np.ndarray
        Simulated time (in seconds).
    tof : Dict[str, float]
        TOF (Turnover Frequency) of gas species (in molecules·s⁻¹·Å⁻²). Example: `KMCOutput.tof['CO2']`.
    total_coverage : np.ndarray
        Total coverage of surface species over time (in %).
    total_coverage_per_site_type : Dict[str, np.ndarray]
        Total coverage of surface species per site type over time (in %). Example: `KMCOutput.total_coverage_per_site_type['top']`.
    total_production : Dict[str, float]
        Total number of gas species produced. Example: `KMCOutput.total_production['CO']`.
    """

    def __init__(self, job_path: str = None, analysis_range: Union[list, None] = None, range_type: str = 'time',
                 weights: Union[str, None] = None, **kwargs):
        """
        Initialize the KMCOutput object by parsing simulation output files.

        Parameters
        ----------
        job_path : str, optional
            The path where the output files are located. (Previously named 'path'.)
            For backward compatibility, you can still pass this value using the keyword 'path'.
        analysis_range : List[float], optional
            A list of two elements `[start_percent, end_percent]` specifying the portion of the entire simulation
            to consider for analysis. The values should be between 0 and 100, representing percentages of the
            total simulated time or the total number of events, depending on `range_type`. For example,
            `[50, 100]` would analyze only the latter half of the simulation. Default is `[0.0, 100.0]`.
        range_type : str, optional
            Determines the dimension used when applying `analysis_range`:
            - `'time'`: The percentages in `analysis_range` refer to segments of the total simulated time.
            - `'nevents'`: The percentages in `analysis_range` refer to segments of the total number of simulated events.
            Default is `'time'`.
        weights : str, optional
            Weights for calculating weighted averages. Possible values are `'time'`, `'nevents'`, or `None`.
            If `None`, all weights are set to 1. Default value is `None`.
        """
        # Support backward compatibility: if job_path is not provided, check for 'path'
        if job_path is None:
            job_path = kwargs.pop('path', None)
        if job_path is None:
            raise TypeError("Missing required argument: job_path (or 'path' for backwards compatibility)")
        self.job_path = job_path

        if analysis_range is None:
            analysis_range = [0.0, 100.0]

        # Parse relevant data from the simulation_input.dat file
        data_simulation = parse_simulation_input_file(
            input_file=f'{self.job_path}/simulation_input.dat')

        self.random_seed = data_simulation['random_seed']
        self.temperature = data_simulation['temperature']
        self.pressure = data_simulation['pressure']
        self.n_gas_species = data_simulation['n_gas_species']
        self.gas_specs_names = data_simulation['gas_specs_names']
        self.gas_molar_fracs = data_simulation['gas_molar_fracs']
        self.n_surf_species = data_simulation['n_surf_species']
        self.surf_specs_names = data_simulation['surf_specs_names']
        self.surf_specs_dent = data_simulation['surf_specs_dent']

        # Parse relevant data from the general_output.txt file
        data_general = parse_general_output_file(
            output_file=f'{self.job_path}/general_output.txt')

        self.n_sites = data_general['n_sites']
        self.area = data_general['area']
        self.site_types = data_general['site_types']
        self.cpu_time = data_general['cpu_time']

        # Parse relevant data from the specnum_output.txt file
        data_specnum, header = parse_specnum_output_file(
            output_file=f'{self.job_path}/specnum_output.txt',
            analysis_range=analysis_range,
            range_type=range_type)

        self.nevents = data_specnum[:, 1]
        self.time = data_specnum[:, 2]
        self.final_time = data_specnum[-1, 2]
        self.energy = data_specnum[:, 4] / self.area  # in eV/Å2

        # If the energy is constant, avoid numerical noise in polyfit by setting slope to zero.
        if np.ptp(self.energy) > 1e-12:
            self.energyslope = abs(np.polyfit(self.nevents, self.energy, 1)[0])  # in eV/Å²/step
        else:
            self.energyslope = 0.0

        self.final_energy = data_specnum[-1, 4] / self.area
        self.av_energy = self.get_average(array=self.energy, weights=weights)

        # Compute production and TOF
        self.production = {}  # in molecules
        self.total_production = {}  # useful when calculating selectivity (i.e., set min_total_production)
        self.tof = {}  # in molecules·s⁻¹·Å⁻²
        for i in range(5 + self.n_surf_species, len(header)):
            gas_spec = header[i]
            self.production[gas_spec] = data_specnum[:, i]
            self.total_production[gas_spec] = data_specnum[-1, i] - data_specnum[0, i]

            gas_data = data_specnum[:, i]
            # Check if production is constant using the peak-to-peak difference.
            if len(gas_data) > 1 and np.ptp(gas_data) > 1e-18:
                slope = np.polyfit(data_specnum[:, 2], gas_data, 1)[0]
                # Force slope to zero if it is within a small tolerance.
                if np.isclose(slope, 0, atol=1e-18):
                    slope = 0.0
                self.tof[gas_spec] = slope / self.area
            else:
                self.tof[gas_spec] = 0.0

        # Compute coverages (per total number of sites)
        dent_sites_map = get_dentate_types(job_path=self.job_path)
        self.coverage = {}
        self.av_coverage = {}
        for i in range(5, 5 + self.n_surf_species):
            surf_spec = header[i]
            dent_sites = dent_sites_map[surf_spec]
            total_dentates = len(dent_sites)
            # total occupied sites = #molecules * #dentates per molecule
            self.coverage[surf_spec] = data_specnum[:, i] * total_dentates / self.n_sites * 100
            self.av_coverage[surf_spec] = self.get_average(
                array=self.coverage[surf_spec], weights=weights
            )
        self.total_coverage = sum(self.coverage.values())
        self.av_total_coverage = min(sum(self.av_coverage.values()), 100)
        self.dominant_ads = max(self.av_coverage, key=self.av_coverage.get)

        # Compute partial coverages (per total number of sites of each site_type)
        self.coverage_per_site_type = {st: {} for st in self.site_types}
        self.av_coverage_per_site_type = {st: {} for st in self.site_types}
        for i in range(5, 5 + self.n_surf_species):
            surf_spec = header[i]
            dent_sites = dent_sites_map[surf_spec]
            for site_type, n_sites_type in self.site_types.items():
                # number of dentates of this species on this site_type
                n_dents_on_type = dent_sites.count(site_type)
                self.coverage_per_site_type[site_type][surf_spec] = (
                    data_specnum[:, i] * n_dents_on_type / n_sites_type * 100
                )
                self.av_coverage_per_site_type[site_type][surf_spec] = self.get_average(
                    array=self.coverage_per_site_type[site_type][surf_spec],
                    weights=weights
                )

        self.total_coverage_per_site_type = {}
        self.av_total_coverage_per_site_type = {}
        self.dominant_ads_per_site_type = {}
        for site_type, cov_dict in self.av_coverage_per_site_type.items():
            if cov_dict:
                self.total_coverage_per_site_type[site_type] = sum(
                    self.coverage_per_site_type[site_type].values()
                )
                self.av_total_coverage_per_site_type[site_type] = min(
                    sum(cov_dict.values()), 100
                )
                self.dominant_ads_per_site_type[site_type] = max(
                    cov_dict, key=cov_dict.get
                )
            else:
                # no adsorption on this site_type
                self.total_coverage_per_site_type[site_type] = np.zeros_like(self.time)
                self.av_total_coverage_per_site_type[site_type] = 0.0
                self.dominant_ads_per_site_type[site_type] = None


    def get_average(self, array, weights):
        """
        Calculate the average of an array with optional weighting.

        Parameters
        ----------
        array : np.ndarray
            The array of values to average.
        weights : str or None
            The weights to apply when calculating the average. Possible values are:
            - `None`: No weighting; all weights are set to 1.
            - `'time'`: Weights based on the differences in simulated time.
            - `'nevents'`: Weights based on the differences in the number of events.

        Returns
        -------
        float
            The calculated average value.
        """
        if weights not in [None, 'time', 'nevents']:
            raise KMCOutputError(f"'weights' must be one of the following: 'none' (default), 'time', or 'nevents'.")

        if len(array) == 1:
            # If the catalyst is poisoned, it could be that the last ∆t is very high and the time window only
            # contains one row. In that case, do not compute the average
            return float(array[0])
        else:
            if weights is None:
                return np.average(array)
            elif weights == 'time':
                return np.average(array[1:], weights=np.diff(self.time))
            elif weights == 'nevents':
                return np.average(array[1:], weights=np.diff(self.nevents))

    @enforce_types
    def get_selectivity(self, main_product: str, side_products: list):
        """
        Calculate the selectivity of the main product over side products.

        Parameters
        ----------
        main_product : str
            Name of the main product.
        side_products : List[str]
            Names of the side products.

        Returns
        -------
        float
            The selectivity of the main product (in %) over the side products.

        Notes
        -----
        The selectivity is calculated as:
            selectivity = (TOF_main_product / (TOF_main_product + sum(TOF_side_products))) * 100

        If the total TOF is zero, the selectivity is returned as NaN.
        """
        selectivity = float('NaN')
        tof_side_products = 0.0
        for side_product in side_products:
            tof_side_products += self.tof[side_product]
        if self.tof[main_product] + tof_side_products != 0:
            selectivity = self.tof[main_product] / (self.tof[main_product] + tof_side_products) * 100
        return selectivity


def get_dentate_types(job_path: str = None, **kwargs):
    """
    Retrieve surface species data including the number of dentates and the associated site types for each surface species.

    Parameters
    ----------
    job_path : str, optional
        The path to the directory containing the simulation input files (simulation_input.dat and energetics_input.dat).
        (Previously named 'path'.) For backward compatibility, you can pass 'path' instead.

    Returns
    -------
    dict
        A dictionary mapping each surface species name to a list of site types indexed by dentate number (1-based).
    """
    import warnings
    # backward compatibility for 'path'
    if job_path is None:
        job_path = kwargs.pop('path', None)
    if job_path is None:
        raise TypeError("Missing required argument: job_path (or 'path' for backwards compatibility)")

    # parse simulation input for species and dentates
    sim_data = parse_simulation_input_file(input_file=f"{job_path}/simulation_input.dat")
    surf_specs = sim_data.get('surf_specs_names', [])
    dent_counts = sim_data.get('surf_specs_dent', [])
    species_dentates = dict(zip(surf_specs, dent_counts))

    # default lattice: every dentate on default site
    if check_default_lattice(job_path=job_path):
        return {spec: ['StTp1'] * species_dentates[spec] for spec in surf_specs}

    # container for global species->dentate mapping
    species_mapping = {}  # species -> {dentate_index: site_type}

    # read energetics_input.dat
    with open(os.path.join(job_path, 'energetics_input.dat'), 'r') as f:
        lines = f.readlines()

    i = 0
    n = len(lines)
    while i < n:
        line = lines[i].strip()
        if line.startswith('cluster'):
            # record cluster identifier for messages
            parts = line.split(maxsplit=1)
            cluster_id = parts[1] if len(parts) > 1 else ''

            # reset block data
            sites_count = None
            state_entries = []
            site_types = []
            i += 1
            # parse cluster block
            while i < n and not lines[i].strip().startswith('end_cluster'):
                text = lines[i].strip()
                if text.startswith('sites'):
                    toks = text.split()
                    if len(toks) >= 2 and toks[1].isdigit():
                        sites_count = int(toks[1])
                if text.startswith('lattice_state'):
                    i += 1
                    count = 0
                    while i < n and count < (sites_count or float('inf')):
                        row = lines[i].strip()
                        if not row or row.startswith('#'):
                            i += 1
                            continue
                        key = row.split()[0]
                        if key in ('site_types', 'graph_multiplicity', 'cluster_eng', 'neighboring', 'angles'):
                            break
                        parts = row.split()
                        if parts and parts[0] == '&':
                            state_entries.append(None)
                        elif len(parts) >= 3 and parts[0].isdigit() and parts[2].isdigit():
                            spec = parts[1]
                            dent = int(parts[2])
                            state_entries.append((spec, dent))
                        i += 1
                        count += 1
                    continue
                if text.startswith('site_types'):
                    toks = text.split()
                    site_types = toks[1:]
                i += 1
            # reached end_cluster
            if sites_count is None:
                raise EnergeticsModelError(f"Missing 'sites' declaration in cluster '{cluster_id}'.")
            if len(site_types) != sites_count:
                raise EnergeticsModelError(
                    f"Cluster '{cluster_id}' declares {sites_count} sites but site_types has {len(site_types)} entries: {site_types}")
            if len(state_entries) != sites_count:
                raise EnergeticsModelError(
                    f"Cluster '{cluster_id}' declares {sites_count} sites but lattice_state provided {len(state_entries)} entries.")

            # build per-cluster mapping: species -> {dentate: site_type}
            cluster_map = {}
            for idx in range(sites_count):
                entry = state_entries[idx]
                stype = site_types[idx]
                if entry is None:
                    continue
                spec, dent = entry
                exp = species_dentates.get(spec)
                if exp is None:
                    raise EnergeticsModelError(
                        f"Species '{spec}' in cluster '{cluster_id}' not declared in simulation_input.dat.")
                if dent < 1 or dent > exp:
                    raise EnergeticsModelError(
                        f"Dentate index {dent} out of range for species '{spec}' in cluster '{cluster_id}' (expected 1..{exp}).")
                cluster_map.setdefault(spec, {})[dent] = stype
            # ensure full dentate coverage per species
            for spec, dent_dict in cluster_map.items():
                if len(dent_dict) != species_dentates[spec]:
                    raise EnergeticsModelError(
                        f"Species '{spec}' in cluster '{cluster_id}' has {species_dentates[spec]} dentates but got {len(dent_dict)} entries.")
            # merge into global mapping with order-flexible consistency
            for spec, dent_dict in cluster_map.items():
                if spec not in species_mapping:
                    species_mapping[spec] = dent_dict.copy()
                else:
                    existing_dict = species_mapping[spec]
                    if existing_dict != dent_dict:
                        # build ordered lists
                        existing_list = [existing_dict[d] for d in sorted(existing_dict)]
                        new_list = [dent_dict[d] for d in sorted(dent_dict)]
                        # same multiset? then warn and keep existing order
                        if sorted(existing_list) == sorted(new_list):
                            warnings.warn(
                                f"In cluster '{cluster_id}', dentates for species '{spec}' appear in {new_list} "
                                f"instead of {existing_list}; keeping original order.")
                        else:
                            raise EnergeticsModelError(
                                f"Inconsistent site types for species '{spec}' in cluster '{cluster_id}': "
                                f"{existing_list} vs {new_list}")
            i += 1  # skip end_cluster
        else:
            i += 1

    # finalize result
    result = {}
    for spec in surf_specs:
        if spec not in species_mapping:
            raise EnergeticsModelError(
                f"Species '{spec}' declared in simulation_input.dat but not found in energetics_input.dat.")
        dent_dict = species_mapping[spec]
        cnt = species_dentates[spec]
        sites = [None] * cnt
        for dent, stype in dent_dict.items():
            sites[dent - 1] = stype
        result[spec] = sites
    return result



def check_default_lattice(job_path: str = None, **kwargs):
    """
    Check whether the simulation uses a default lattice configuration.

    Parameters
    ----------
    job_path : str, optional
        The path to the directory containing the `lattice_input.dat` file.
        (Previously named 'path'.) For backward compatibility, you can pass 'path' instead.

    Returns
    -------
    bool
        True if the `lattice_input.dat` file indicates a default lattice configuration, False otherwise.
    """

    if job_path is None:
        job_path = kwargs.pop('path', None)
    if job_path is None:
        raise TypeError("Missing required argument: job_path (or 'path' for backwards compatibility)")

    with open(os.path.join(job_path, 'lattice_input.dat'), 'r') as file:
        for line in file:
            if 'lattice' in line and 'default_choice' in line:
                return True
    return False



from pathlib import Path
import numpy as np
import pandas as pd
from zacrostools.general_output import parse_general_output_file
from typing import Union, List, Tuple, Optional, Dict
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import warnings


def parse_procstat_output_file(procstat_output_path: Union[str, Path],
                               analysis_range: List[float],
                               range_type: str) -> Tuple[pd.DataFrame, float, int, float]:
    """
    Parses the procstat_output.txt file and extracts raw occurrence data within a specified window.

    Parameters
    ----------
    procstat_output_path : Union[str, Path]
        Path to the **folder** that contains 'procstat_output.txt'.
    analysis_range : List[float], optional
        A list of two elements `[start_percent, end_percent]` specifying the portion of the entire simulation
        to consider for analysis. The values should be between 0 and 100, representing percentages of the
        total simulated time or the total number of events, depending on `range_type`. For example,
        `[50, 100]` would analyze only the latter half of the simulation. Default is `[0.0, 100.0]`.
    range_type : str, optional
        Determines the dimension used when applying `analysis_range`:
        - `'time'`: The percentages in `analysis_range` refer to segments of the total simulated time.
        - `'nevents'`: The percentages in `analysis_range` refer to segments of the total number of simulated events.
        Default is `'time'`.

    Returns
    -------
    Tuple[pd.DataFrame, float, float]
        A tuple containing:
        - A DataFrame where each row corresponds to an elementary step (without _fwd/_rev suffix)
          and the columns are:
            - noccur_fwd
            - noccur_rev
            - noccur_net
        - delta_time : float
            The time interval corresponding to the selected window.
        - delta_events : int
            The number of events corresponding to the selected window.
        - area : float
            The area extracted from general_output.txt.
    """

    procstat_output_path = Path(procstat_output_path)
    output_file = procstat_output_path / "procstat_output.txt"
    if not output_file.is_file():
        raise FileNotFoundError(f"Output file '{output_file}' does not exist.")

    # Read entire file
    with output_file.open('r') as f:
        lines = [line.strip() for line in f if line.strip()]

    if not lines:
        raise ValueError("The output file is empty.")

    header_line = lines[0]
    headers = header_line.split()

    if headers[0].lower() != 'overall':
        raise ValueError("The first header in procstat_output.txt must be 'Overall'.")

    # headers looks like: ['Overall', 'elemstep1_fwd', 'elemstep1_rev', 'elemstep2_fwd', 'elemstep2_rev', ...]
    # Group them into events
    event_names = []
    event_fwd_indices = []
    event_rev_indices = []

    i = 1  # Skip the "Overall" column
    while i < len(headers):
        col_name = headers[i]
        if col_name.endswith('_fwd'):
            base_name = col_name.replace('_fwd', '')
            # Check if the next header exists and is the reverse counterpart
            if i + 1 < len(headers) and headers[i + 1] == f"{base_name}_rev":
                event_names.append(base_name)
                event_fwd_indices.append(i)
                event_rev_indices.append(i + 1)
                i += 2  # Skip the next column as well
            else:
                # Treat as an irreversible step even though labeled _fwd
                event_names.append(base_name)
                event_fwd_indices.append(i)
                event_rev_indices.append(None)
                i += 1
        elif col_name.endswith('_rev'):
            # Option: raise an error because a reverse column without a forward is unexpected
            raise ValueError(f"Found a reverse column '{col_name}' without a matching forward column.")
        else:
            # Irreversible step with no suffix
            event_names.append(col_name)
            event_fwd_indices.append(i)
            event_rev_indices.append(None)
            i += 1

    # From line 1 onwards, the file has appended blocks of 3 lines each time the file is updated:
    # 1) "configuration int1 int2 real1"
    #    - int1: configuration count
    #    - int2: number of events so far
    #    - real1: current simulation time
    # 2) line with average waiting times (ignore)
    # 3) line with cumulative event counts:
    #    - int3: total number of events (should match int2)
    #    - int4: occurrences of elemstep1_fwd
    #    - int5: occurrences of elemstep1_rev
    #    etc.

    # We need to extract arrays for:
    #   configurations, nevents, time, and cumulative counts for each event

    # Skip the header line
    data_lines = lines[1:]

    # Every block of interest has 3 lines:
    # pattern:
    #   line0: configuration ...
    #   line1: (ignored waiting times)
    #   line2: cumulative counts line
    # So we need to parse these in groups of 3.
    if len(data_lines) % 3 != 0:
        # The file might end in partial blocks; handle gracefully or raise an error
        # We'll just use complete blocks
        trunc_len = (len(data_lines) // 3) * 3
        data_lines = data_lines[:trunc_len]

    # Initialize lists to store data
    configuration_counts = []
    total_events = []
    times = []
    cumulative_counts = []

    for i in range(0, len(data_lines), 3):
        config_line = data_lines[i]
        # ignored_line = data_lines[i+1]  # Ignored
        counts_line = data_lines[i + 2]

        # Parse configuration line
        # Example: "configuration 1 100 2.5"
        config_parts = config_line.split()
        if config_parts[0].lower() != 'configuration':
            raise ValueError("Expected a 'configuration' line.")
        if len(config_parts) < 4:
            raise ValueError("Configuration line does not have enough elements.")
        try:
            conf_count = int(config_parts[1])
            n_events = int(config_parts[2])
            t_sim = float(config_parts[3])
        except ValueError as e:
            raise ValueError(f"Error parsing configuration line: {e}")

        configuration_counts.append(conf_count)
        total_events.append(n_events)
        times.append(t_sim)

        # Parse counts line
        # This should have the same number of columns as headers
        # Format: int3 int4 int5 ...
        cparts = counts_line.split()
        if len(cparts) != len(headers):
            raise ValueError("Counts line does not match the number of header columns.")
        try:
            cparts_int = list(map(int, cparts))
        except ValueError as e:
            raise ValueError(f"Error parsing counts line: {e}")
        cumulative_counts.append(cparts_int)

    total_events = np.array(total_events, dtype=int)
    times = np.array(times, dtype=float)
    cumulative_counts = np.array(cumulative_counts, dtype=int)  # shape: (Nblocks, Ncolumns_in_header)

    if len(times) == 0:
        raise ValueError("No data found in procstat_output.txt.")

    finaltime = times[-1]
    final_nevents = total_events[-1]

    start_percent = analysis_range[0]
    end_percent = analysis_range[1]

    if range_type not in ['time', 'nevents']:
        raise ValueError("range_type must be either 'time' or 'nevents'.")

    if range_type == 'time':
        # Convert percent of time to actual time
        start_time = (start_percent / 100.0) * finaltime
        end_time = (end_percent / 100.0) * finaltime

        # Find indices that bracket start_time and end_time
        # We want the block that is just before or at start_time and the block at or just after end_time
        start_idx = np.searchsorted(times, start_time, side='right') - 1
        if start_idx < 0:
            start_idx = 0
        end_idx = np.searchsorted(times, end_time, side='right') - 1
        if end_idx < 0:
            end_idx = 0
        end_idx = min(end_idx, len(times) - 1)

    else:  # range_type == 'nevents'
        # Convert percent of events to actual number of events
        start_nevents = (start_percent / 100.0) * final_nevents
        end_nevents = (end_percent / 100.0) * final_nevents

        start_idx = np.searchsorted(total_events, start_nevents, side='right') - 1
        if start_idx < 0:
            start_idx = 0
        end_idx = np.searchsorted(total_events, end_nevents, side='right') - 1
        if end_idx < 0:
            end_idx = 0
        end_idx = min(end_idx, len(total_events) - 1)

    # If start_idx and end_idx are the same, it means zero-length window. Handle gracefully:
    if start_idx == end_idx:
        # This might indicate that the chosen window doesn't contain any interval
        # We can either raise an error or return an empty DataFrame
        return pd.DataFrame(columns=['noccur_fwd', 'noccur_rev', 'noccur_net']), 0.0, 0, 1.0

    # Calculate the differences in times and events
    delta_time = times[end_idx] - times[start_idx]
    delta_events = total_events[end_idx] - total_events[start_idx]

    # If delta_time is zero, that means no time interval. Avoid division by zero:
    if delta_time <= 0.0:
        return pd.DataFrame(columns=['noccur_fwd', 'noccur_rev', 'noccur_net']), 0.0, int(delta_events), 1.0

    # Compute differences in counts for each event
    # cumulative_counts[end_idx] - cumulative_counts[start_idx]
    delta_counts = cumulative_counts[end_idx] - cumulative_counts[start_idx]

    # Create lists to store counts
    noccur_fwd_list = []
    noccur_rev_list = []

    for fwd_idx, rev_idx in zip(event_fwd_indices, event_rev_indices):
        fwd_count = delta_counts[fwd_idx]
        if rev_idx is None:
            rev_count = 0  # Irreversible: no reverse events
        else:
            rev_count = delta_counts[rev_idx]
        noccur_fwd_list.append(fwd_count)
        noccur_rev_list.append(rev_count)

    # Compute net occurrences
    noccur_net = np.array(noccur_fwd_list) - np.array(noccur_rev_list)

    # Create a DataFrame with raw counts
    df = pd.DataFrame({
        'noccur_fwd': noccur_fwd_list,
        'noccur_rev': noccur_rev_list,
        'noccur_net': noccur_net
    }, index=event_names)

    # Parse general_output.txt to get area
    # Assuming general_output.txt is in the same directory as procstat_output.txt
    general_output_path = output_file.parent / "general_output.txt"
    try:
        general_data = parse_general_output_file(general_output_path)
        area = general_data.get('area', 1.0)  # Default to 1.0 if not found
    except Exception as e:
        raise ValueError(f"Error parsing 'general_output.txt': {e}")


    return df, float(delta_time), int(delta_events), float(area)


def plot_event_frequency(
        ax: plt.Axes,
        simulation_path: Union[str, Path],
        analysis_range: List[float],
        range_type: str,
        elementary_steps: Optional[List[str]] = None,
        hide_zero_events: bool = False,
        grouping: Optional[Dict[str, List[str]]] = None,
        show_minimum: bool = True,
        print_pe_ratio: bool = False,
        print_freq: bool = False
) -> plt.Axes:
    """
    Parse the procstat_output.txt file from a given simulation path and produce a
    horizontal bar plot of event frequencies on the given Axes object, with optional grouping of elementary steps.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axes on which to plot the data.
    simulation_path : Union[str, Path]
        The path to the simulation directory containing procstat_output.txt and general_output.txt.
    analysis_range : List[float]
        [start_percent, end_percent] defining the slicing window in terms of simulation time or events.
    range_type : str
        'time' or 'nevents', defining whether the slicing window is based on simulation time or number of events.
    elementary_steps : Optional[List[str]], default None
        If provided, only those steps are included in the plot. Otherwise, all steps are included.
        This will be performed before grouping and affects the original elementary step names before grouping.
    hide_zero_events : bool, default False
        If True, steps with zero occurrences in the fwd and rev directions are omitted from the plot.
    grouping : Optional[Dict[str, List[str]]], default None
        A dictionary defining groups of elementary steps. Each key is the name of the grouped step,
        and each value is a list of elementary steps to be grouped together.
    show_minimum: if True, draw the vertical line for the minimum event frequency.
    print_pe_ratio: if True, print the P/E ratio for each step.
    print_freq: if True, print the computed event frequencies.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes with the plot drawn on it.
    """

    simulation_path = Path(simulation_path)

    df_counts, delta_time, delta_events, area = parse_procstat_output_file(
        procstat_output_path=simulation_path,
        analysis_range=analysis_range,
        range_type=range_type
    )

    if df_counts.empty:
        raise ValueError(f"No steps have occurred.")

    if elementary_steps:
        # Remove rows from df_counts that are not in elementary_steps
        df_counts_filtered = df_counts.loc[df_counts.index.isin(elementary_steps)].copy()
        # Identify steps in elementary_steps not present in df_counts and trigger a warning for each missing step
        missing_steps = set(elementary_steps) - set(df_counts.index)
        for step in missing_steps:
            warnings.warn(f"Step '{step}' is in elementary_steps but not in df_counts.")
    else:
        df_counts_filtered = df_counts.copy()

    if grouping:
        # Validate the grouping dictionary
        valid_grouping = validate_grouping(grouping, df_counts_filtered)

        # Create a step to group mapping
        step_to_group = {}
        for group, steps in valid_grouping.items():
            for step in steps:
                step_to_group[step] = group

        # Function to map each step to its group or itself
        def map_step(step):
            return step_to_group.get(step, step)

        # Apply the mapping to the index to create new grouping
        new_index = df_counts_filtered.index.to_series().apply(map_step)

        # Group by the new index and sum the counts
        df_counts_grouped = df_counts_filtered.groupby(new_index).sum()

        # Now, to preserve the original order:
        # For each group or individual step, find the first occurrence in original_order
        original_order = df_counts.index.to_list()
        group_order = {}
        for idx, name in enumerate(original_order):
            if name in step_to_group:
                group = step_to_group[name]
                if group not in group_order:
                    group_order[group] = idx
            else:
                group_order[name] = idx
        df_counts_grouped = df_counts_grouped.copy()
        sort_keys = []
        for group in df_counts_grouped.index:
            if group in group_order:
                sort_keys.append(group_order[group])
            else:
                # If group not in group_order, assign a large number to push it to the end
                sort_keys.append(len(original_order))
        df_counts_grouped['sort_key'] = sort_keys
        df_counts_grouped = df_counts_grouped.sort_values('sort_key').drop(columns='sort_key')
    else:
        df_counts_grouped = df_counts_filtered.copy()

    if hide_zero_events:
        condition = (df_counts_grouped['noccur_fwd'] != 0) | (df_counts_grouped['noccur_rev'] != 0)
        df_counts_grouped = df_counts_grouped[condition]

    # print P/E ratios
    if print_pe_ratio:
        for step, row in df_counts_grouped.iterrows():
            n_fwd = row['noccur_fwd']
            n_rev = row['noccur_rev']
            total = n_fwd + n_rev
            pe = n_fwd / total if total > 0 else np.nan
            print(f"Step {step}: P/E ratio = {pe:.3f}")

    # Calculate frequencies: frequencies = counts / (delta_time * area)
    df_freq = df_counts_grouped.copy()
    df_freq['eventfreq_fwd'] = df_freq['noccur_fwd'] / (delta_time * area)
    df_freq['eventfreq_rev'] = df_freq['noccur_rev'] / (delta_time * area)
    df_freq['eventfreq_net'] = df_freq['noccur_net'] / (delta_time * area)

    # Handle any potential infinities or NaNs resulting from division
    df_freq.replace([np.inf, -np.inf], np.nan, inplace=True)
    df_freq.fillna(0.0, inplace=True)

    # print frequencies
    if print_freq:
        for step, row in df_freq.iterrows():
            fwd = row['eventfreq_fwd']
            rev = row['eventfreq_rev']
            net = row['eventfreq_net']
            print(f"Step {step}: freq_fwd={fwd:.3e}, freq_rev={rev:.3e}, freq_net={net:.3e}")

    # Define colors for the barplot
    color_fwd = '#4169e1'      # Royal Blue
    color_rev = '#daa520'      # Goldenrod
    color_net_pos = '#32cd32'  # Lime Green
    color_net_neg = '#dc143c'  # Crimson

    bar_height = 0.22

    offset_map = {
        'Forward': bar_height * (-1.1),
        'Reverse': 0.0,
        'Net (+)': bar_height * 1.1,
        'Net (-)': bar_height * 1.1
    }

    y_tick_positions = []
    y_tick_labels = []

    for i, step in enumerate(df_freq.index):
        fwd = df_freq.loc[step, 'eventfreq_fwd']
        rev = df_freq.loc[step, 'eventfreq_rev']
        net = df_freq.loc[step, 'eventfreq_net']
        # Forward
        if fwd > 0:
            ax.barh(y=i + offset_map['Forward'], width=fwd, height=bar_height,
                    color=color_fwd, align='center', edgecolor='black', linewidth=0.5, zorder=3)
        # Reverse
        if rev > 0:
            ax.barh(y=i + offset_map['Reverse'], width=rev, height=bar_height,
                    color=color_rev, align='center', edgecolor='black', linewidth=0.5, zorder=3)
        # Net (+)
        if net > 0:
            ax.barh(y=i + offset_map['Net (+)'], width=net, height=bar_height,
                    color=color_net_pos, align='center', edgecolor='black', linewidth=0.5, zorder=3)
        # Net (-)
        if net < 0:
            abs_net = abs(net)
            ax.barh(y=i + offset_map['Net (-)'], width=abs_net, height=bar_height,
                    color=color_net_neg, align='center', edgecolor='black', linewidth=0.5, zorder=3)

        y_tick_positions.append(i)
        y_tick_labels.append(step)

    ax.set_yticks(y_tick_positions)
    ax.set_yticklabels(y_tick_labels)
    ax.set_xscale('log')
    ax.set_xlabel(r'Event frequency ($\mathrm{s^{-1}\,\AA^{-2}}$)', fontsize=14)
    ax.set_ylabel('Elementary step', fontsize=14)

    # Vertical line for minimum frequency
    min_eventfreq = 1.0 / (delta_time * area)
    if show_minimum and np.isfinite(min_eventfreq) and min_eventfreq > 0:
        ax.axvline(
            x=min_eventfreq,
            color='black',
            linestyle='-',
            linewidth=1.5,
            zorder=4,
            label='Min. event frequency'
        )

    legend_handles = [
        Line2D([0], [0], color=color_fwd, lw=10, label='Forward'),
        Line2D([0], [0], color=color_rev, lw=10, label='Reverse'),
        Line2D([0], [0], color=color_net_pos, lw=10, label='Net (+)'),
        Line2D([0], [0], color=color_net_neg, lw=10, label='Net (-)')
    ]

    # Add the min_eventfreq line to the legend if it was plotted,
    if np.isfinite(min_eventfreq) and min_eventfreq > 0:
        legend_handles.append(Line2D([0], [0], color='black', linestyle='-', lw=2, label='Min. event frequency'))
        legend_labels = ['Forward', 'Reverse', 'Net (+)', 'Net (-)', 'Min. event frequency']
    else:
        legend_labels = ['Forward', 'Reverse', 'Net (+)', 'Net (-)']

    ax.legend(handles=legend_handles, labels=legend_labels, loc='best')
    ax.grid(True, which='both', axis='x', linestyle='--', linewidth=0.5, alpha=0.7)
    ax.invert_yaxis()
    return ax


def validate_grouping(grouping, df_counts_filtered):
    """
    Validates the grouping dictionary.

    Parameters:
    - grouping (dict): Dictionary defining groups and their corresponding steps.
    - df_counts_filtered (pd.DataFrame): DataFrame containing the steps to be grouped.

    Returns:
    - valid_grouping (dict): A cleaned version of the grouping dictionary with invalid steps removed.
    """

    # 1. Ensure the same elementary step is not in multiple groups
    all_steps = []
    for group, steps in grouping.items():
        if not isinstance(steps, list):
            raise TypeError(f"Steps for group '{group}' must be provided as a list.")
        all_steps.extend(steps)

    duplicate_steps = set([step for step in all_steps if all_steps.count(step) > 1])
    if duplicate_steps:
        raise ValueError(f"The following steps are assigned to multiple groups: {', '.join(duplicate_steps)}")

    # 2. Check if steps in grouping are present in df_counts_filtered.index
    df_steps_set = set(df_counts_filtered.index)
    grouping_steps_set = set(all_steps)
    missing_steps = grouping_steps_set - df_steps_set
    if missing_steps:
        warnings.warn(
            f"The following steps specified in grouping are not present in df_counts_filtered and will be ignored: {', '.join(missing_steps)}")
        # Remove missing steps from grouping
        for group in grouping:
            original_length = len(grouping[group])
            grouping[group] = [step for step in grouping[group] if step in df_steps_set]
            if len(grouping[group]) < original_length:
                warnings.warn(
                    f"In group '{group}', {original_length - len(grouping[group])} step(s) were removed because they are not present in df_counts_filtered.")

    return grouping



from random import randint
from typing import Union, Optional
from pathlib import Path

from zacrostools.header import write_header
from zacrostools.lattice_model import LatticeModel
from zacrostools.energetics_model import EnergeticsModel
from zacrostools.reaction_model import ReactionModel
from zacrostools.gas_model import GasModel
from zacrostools.custom_exceptions import KMCModelError
from zacrostools.custom_exceptions import enforce_types


class KMCModel:
    """
    Represents a Kinetic Monte Carlo (KMC) model.

    Parameters
    ----------
    gas_model : GasModel
        An instance containing information about the gas molecules.
    reaction_model : ReactionModel
        An instance containing information about the reaction model.
    energetics_model : EnergeticsModel
        An instance containing information about the energetic model.
    lattice_model : LatticeModel
        An instance containing information about the lattice model.
    """

    @enforce_types
    def __init__(self,
                 gas_model: GasModel,
                 reaction_model: ReactionModel,
                 energetics_model: EnergeticsModel,
                 lattice_model: LatticeModel):
        self.job_dir: Optional[Path] = None
        self.gas_model = gas_model
        self.reaction_model = reaction_model
        self.energetics_model = energetics_model
        self.lattice_model = lattice_model
        self.check_errors()

    def check_errors(self):
        """
        Check for data consistency after initialization.

        Raises
        ------
        KMCModelError
            If there are inconsistencies in the model configurations.
        """
        if self.lattice_model.lattice_type == 'default_choice':
            if 'site_types' in self.reaction_model.df.columns:
                raise KMCModelError("Remove 'site_types' from the reaction model when using a default lattice.")
            if 'site_types' in self.energetics_model.df.columns:
                raise KMCModelError("Remove 'site_types' from the energetic model when using a default lattice.")
        else:
            if 'site_types' not in self.reaction_model.df.columns:
                raise KMCModelError("'site_types' are missing in the reaction model.")
            if 'site_types' not in self.energetics_model.df.columns:
                raise KMCModelError("'site_types' are missing in the energetic model.")

    @enforce_types
    def create_job_dir(self,
                       job_path: str,
                       temperature: Union[float, int],
                       pressure: dict,
                       reporting_scheme: Optional[dict] = None,
                       stopping_criteria: Optional[dict] = None,
                       manual_scaling: Optional[dict] = None,
                       stiffness_scaling_algorithm: Optional[str] = None,
                       stiffness_scalable_steps: Union[list, str, None] = None,
                       stiffness_scalable_symmetric_steps: Optional[list] = None,
                       stiffness_scaling_tags: Optional[dict] = None,
                       additional_keywords: Optional[list] = None,
                       sig_figs_energies: int = 8,
                       sig_figs_pe: int = 8,
                       sig_figs_lattice: int = 8,
                       random_seed: Optional[int] = None,
                       version: Union[float, int] = 5.0):
        """
        Create a job directory and write the necessary input files for the KMC simulation.

        Parameters
        ----------
        job_path : str
            The path for the job directory where input files will be written.
        temperature : float or int
            Reaction temperature (in K).
        pressure : dict
            Partial pressures of all gas species (in bar), e.g., {'CO': 1.0, 'O2': 0.001}.
        reporting_scheme : dict, optional
            Reporting scheme in Zacros format. Must contain the following keys:
            'snapshots', 'process_statistics', and 'species_numbers'.
            Default is {'snapshots': 'on event 10000',
                        'process_statistics': 'on event 10000',
                        'species_numbers': 'on event 10000'}.
        stopping_criteria : dict, optional
            Stopping criteria in Zacros format. Must contain the following keys:
            'max_steps', 'max_time', and 'wall_time'.
            Default is {'max_steps': 'infinity',
                        'max_time': 'infinity',
                        'wall_time': 86400}.
        manual_scaling : dict, optional
            Dictionary mapping step names to their corresponding manual scaling factors, e.g.,
            {'CO_diffusion': 1.0e-1, 'O_diffusion': 1.0e-2}. Default is {}.
        stiffness_scaling_algorithm : str, optional
            Algorithm used for stiffness scaling. Allowed values are None (default), 'legacy', or 'prats2024'.
        stiffness_scalable_steps : list of str or 'all', optional
            Steps that will be marked as 'stiffness_scalable' in mechanism_input.dat.
            Can be provided as a list of step names or the string 'all' to indicate that all steps
            (except those specified in stiffness_scalable_symmetric_steps) are stiffness scalable.
            Default is [].
        stiffness_scalable_symmetric_steps : list of str, optional
            Steps that will be marked as 'stiffness_scalable_symmetric' in mechanism_input.dat.
            Default is [].
        stiffness_scaling_tags : dict, optional
            Keywords controlling the dynamic stiffness scaling algorithm and their corresponding values, e.g.,
            {'check_every': 500, 'min_separation': 400.0, ...}.
            The correct types are: integer for 'check_every' and 'min_noccur', and float for all others.
            Default is {}
        additional_keywords : list of str, optional
            Additional keywords to append to simulation_input.dat that are currently not supported by zacrostools,
            e.g., ['event_report on']. Each keyword will be written on its own line.
            Default is None.
        sig_figs_energies : int, optional
            Number of significant figures for energy values in input files.
            Default is 8.
        sig_figs_pe : int, optional
            Number of significant figures for pre-exponential factors in mechanism_input.dat.
            Default is 8.
        sig_figs_lattice : int, optional
            Number of significant figures for coordinates in lattice_input.dat.
            Default is 8.
        random_seed : int, optional
            The seed for the random number generator. If not specified, a random seed will be generated.
            Default is None.
        version : float or int, optional
            The Zacros version. Can be a single integer (e.g. 4) or float (e.g. 4.2 or 5.1).
        """
        # Parse and validate parameters
        parsed_params = self._parse_parameters(
            reporting_scheme=reporting_scheme,
            stopping_criteria=stopping_criteria,
            manual_scaling=manual_scaling,
            stiffness_scaling_algorithm=stiffness_scaling_algorithm,
            stiffness_scalable_steps=stiffness_scalable_steps,
            stiffness_scalable_symmetric_steps=stiffness_scalable_symmetric_steps,
            stiffness_scaling_tags=stiffness_scaling_tags,
            version=version
        )

        # Unpack parsed parameters
        reporting_scheme = parsed_params['reporting_scheme']
        stopping_criteria = parsed_params['stopping_criteria']
        manual_scaling = parsed_params['manual_scaling']
        stiffness_scaling_algorithm = parsed_params['stiffness_scaling_algorithm']
        stiffness_scalable_steps = parsed_params['stiffness_scalable_steps']
        stiffness_scalable_symmetric_steps = parsed_params['stiffness_scalable_symmetric_steps']
        stiffness_scaling_tags = parsed_params['stiffness_scaling_tags']

        additional_keywords = additional_keywords or []

        self.job_dir = Path(job_path)
        if not self.job_dir.exists():
            self.job_dir.mkdir(parents=True, exist_ok=True)
            self.write_simulation_input(
                temperature=temperature,
                pressure=pressure,
                reporting_scheme=reporting_scheme,
                stopping_criteria=stopping_criteria,
                stiffness_scaling_algorithm=stiffness_scaling_algorithm,
                stiffness_scalable_steps=stiffness_scalable_steps,
                stiffness_scalable_symmetric_steps=stiffness_scalable_symmetric_steps,
                stiffness_scaling_tags=stiffness_scaling_tags,
                additional_keywords=additional_keywords,
                sig_figs_energies=sig_figs_energies,
                random_seed=random_seed,
                version=version)
            self.reaction_model.write_mechanism_input(
                output_dir=self.job_dir,
                temperature=temperature,
                gas_model=self.gas_model,
                manual_scaling=manual_scaling,
                stiffness_scalable_steps=stiffness_scalable_steps,
                stiffness_scalable_symmetric_steps=stiffness_scalable_symmetric_steps,
                sig_figs_energies=sig_figs_energies,
                sig_figs_pe=sig_figs_pe)
            self.energetics_model.write_energetics_input(
                output_dir=self.job_dir,
                sig_figs_energies=sig_figs_energies)
            self.lattice_model.write_lattice_input(
                output_dir=self.job_dir,
                sig_figs=sig_figs_lattice)
        else:
            print(f'{self.job_dir} already exists (nothing done)')

    def _parse_parameters(self,
                          reporting_scheme,
                          stopping_criteria,
                          manual_scaling,
                          stiffness_scaling_algorithm,
                          stiffness_scalable_steps,
                          stiffness_scalable_symmetric_steps,
                          stiffness_scaling_tags,
                          version):
        """
        Parse and validate the parameters provided to create_job_dir.
        """
        # [Reporting scheme and stopping criteria processing as before...]
        allowed_reporting_keys = {'snapshots', 'process_statistics', 'species_numbers'}
        default_reporting_scheme = {
            'snapshots': 'on event 10000',
            'process_statistics': 'on event 10000',
            'species_numbers': 'on event 10000'
        }
        if reporting_scheme is None:
            reporting_scheme = default_reporting_scheme
        else:
            reporting_scheme = {key: reporting_scheme.get(key, default_reporting_scheme[key])
                                for key in allowed_reporting_keys}

        allowed_stopping_keys = {'max_steps', 'max_time', 'wall_time'}
        default_stopping_criteria = {
            'max_steps': 'infinity',
            'max_time': float(1.0e+10),
            'wall_time': 86400
        }
        if stopping_criteria is None:
            stopping_criteria = default_stopping_criteria
        else:
            stopping_criteria = {key: stopping_criteria.get(key, default_stopping_criteria[key])
                                 for key in allowed_stopping_keys}

        # Validate 'max_steps', 'max_time' and 'wall_time'
        if 'max_steps' in stopping_criteria:
            ms = stopping_criteria['max_steps']
            if ms != 'infinity' and not isinstance(ms, int):
                raise KMCModelError(
                    f"'max_steps' must be either 'infinity' or an integer, got {ms}."
                )
        if 'max_time' in stopping_criteria:
            mt = stopping_criteria['max_time']
            if mt != 'infinity' and not isinstance(mt, float):
                raise KMCModelError(
                    f"'max_time' must be either 'infinity' or a float, got {mt}."
                )
        if 'wall_time' in stopping_criteria:
            wt = stopping_criteria['wall_time']
            if not isinstance(wt, int):
                raise KMCModelError(
                    f"'wall_time' must be an integer, got {wt}."
                )

        # Manual scaling
        if manual_scaling is None:
            manual_scaling = {}

        # Handle stiffness_scalable_steps if set to 'all'
        if isinstance(stiffness_scalable_steps, str):
            if stiffness_scalable_steps.lower() == 'all':
                # Assign all step names from the reaction model except those in stiffness_scalable_symmetric_steps
                all_steps = list(self.reaction_model.df.index)
                stiffness_scalable_steps = [step for step in all_steps
                                            if step not in stiffness_scalable_symmetric_steps]
            else:
                raise KMCModelError(
                    "Invalid value for stiffness_scalable_steps: if provided as a string, only 'all' is allowed."
                )

        # Version-specific stiffness scaling validation
        if stiffness_scalable_steps is None:
            stiffness_scalable_steps = []
        if stiffness_scalable_symmetric_steps is None:
            stiffness_scalable_symmetric_steps = []
        if stiffness_scaling_tags is None:
            stiffness_scaling_tags = {}

        if float(version) < 3.0:
            if (stiffness_scaling_algorithm is not None or
                    (stiffness_scalable_steps not in (None, []) and len(stiffness_scalable_steps) > 0) or
                    (stiffness_scalable_symmetric_steps not in (None, []) and len(
                        stiffness_scalable_symmetric_steps) > 0) or
                    (stiffness_scaling_tags not in (None, {}) and len(stiffness_scaling_tags) > 0)):
                raise KMCModelError("Stiffness scaling is not implemented for Zacros versions lower than 3.")
        elif 3.0 <= float(version) < 5.0:
            if stiffness_scaling_algorithm is not None:
                print(
                    "Warning: 'stiffness_scaling_algorithm' parameter is not allowed for Zacros versions lower than 5. Ignoring it.")
                stiffness_scaling_algorithm = None
            if stiffness_scalable_symmetric_steps not in (None, []) and len(stiffness_scalable_symmetric_steps) > 0:
                print(
                    "Warning: 'stiffness_scalable_symmetric_steps' parameter is not allowed for Zacros versions lower than 5. Ignoring it.")
                stiffness_scalable_symmetric_steps = []
        else:
            allowed_scaling_algorithms = {'legacy', 'prats2024'}
            if stiffness_scaling_algorithm is not None:
                if stiffness_scaling_algorithm not in allowed_scaling_algorithms:
                    raise KMCModelError(
                        f"Invalid stiffness_scaling_algorithm '{stiffness_scaling_algorithm}'. "
                        f"Allowed values are 'legacy' or 'prats2024'."
                    )
            if stiffness_scaling_algorithm is None:
                if stiffness_scalable_steps or stiffness_scalable_symmetric_steps or stiffness_scaling_tags:
                    stiffness_scaling_algorithm = 'legacy'
                else:
                    stiffness_scaling_algorithm = None

        if float(version) >= 5.0:
            if stiffness_scaling_algorithm in {'legacy', 'prats2024'}:
                if not stiffness_scalable_steps and not stiffness_scalable_symmetric_steps:
                    raise KMCModelError(
                        "stiffness_scaling_algorithm selected but no steps are stiffness scalable."
                    )
        # Validate stiffness scaling tags
        if float(version) >= 5.0:
            if stiffness_scaling_algorithm == 'legacy':
                allowed_tags = {
                    'check_every',
                    'min_separation',
                    'max_separation',
                    'max_qequil_separation',
                    'tol_part_equil_ratio',
                    'stiffn_coeff_threshold',
                    'scaling_factor'
                }
            elif stiffness_scaling_algorithm == 'prats2024':
                allowed_tags = {
                    'check_every',
                    'min_separation',
                    'max_separation',
                    'tol_part_equil_ratio',
                    'upscaling_factor',
                    'upscaling_limit',
                    'downscaling_limit',
                    'min_noccur'
                }
            else:
                allowed_tags = set()
        else:
            allowed_tags = {
                'check_every',
                'min_separation',
                'max_separation',
                'max_qequil_separation',
                'tol_part_equil_ratio',
                'stiffn_coeff_threshold',
                'scaling_factor'
            }

        if stiffness_scaling_tags:
            invalid_tags = set(stiffness_scaling_tags.keys()) - allowed_tags
            if invalid_tags:
                raise KMCModelError(
                    f"Invalid stiffness_scaling_tags keys for Zacros version {version}: "
                    f"{invalid_tags}. Allowed keys are: {allowed_tags}."
                )
            for tag, value in stiffness_scaling_tags.items():
                if tag in ['check_every', 'min_noccur']:
                    if not isinstance(value, int):
                        raise KMCModelError(
                            f"Invalid type for stiffness_scaling_tags '{tag}': expected int, got {type(value).__name__}."
                        )
                else:
                    if not isinstance(value, float):
                        raise KMCModelError(
                            f"Invalid type for stiffness_scaling_tags '{tag}': expected float, got {type(value).__name__}."
                        )

        return {
            'reporting_scheme': reporting_scheme,
            'stopping_criteria': stopping_criteria,
            'manual_scaling': manual_scaling,
            'stiffness_scaling_algorithm': stiffness_scaling_algorithm,
            'stiffness_scalable_steps': stiffness_scalable_steps,
            'stiffness_scalable_symmetric_steps': stiffness_scalable_symmetric_steps,
            'stiffness_scaling_tags': stiffness_scaling_tags
        }

    def write_simulation_input(self,
                               temperature,
                               pressure,
                               reporting_scheme,
                               stopping_criteria,
                               stiffness_scaling_algorithm,
                               stiffness_scalable_steps,
                               stiffness_scalable_symmetric_steps,
                               stiffness_scaling_tags,
                               additional_keywords=None,
                               sig_figs_energies=None,
                               random_seed=None,
                               version=None):
        """
        Writes the simulation_input.dat file.
        """
        gas_specs_names = list(self.gas_model.df.index)
        surf_specs = self.get_surf_specs()
        write_header(f"{self.job_dir}/simulation_input.dat")
        try:
            with open(f"{self.job_dir}/simulation_input.dat", 'a') as infile:
                # Handle random seed
                if random_seed is None:
                    infile.write('random_seed\t'.expandtabs(26) + str(randint(100000, 999999)) + '\n')
                else:
                    infile.write('random_seed\t'.expandtabs(26) + str(random_seed) + '\n')

                # Write temperature
                infile.write('temperature\t'.expandtabs(26) + str(float(temperature)) + '\n')

                # Write total pressure
                p_tot = sum(pressure.values())
                infile.write('pressure\t'.expandtabs(26) + str(float(p_tot)) + '\n')

                # Write number of gas species and their names
                infile.write('n_gas_species\t'.expandtabs(26) + str(len(gas_specs_names)) + '\n')
                infile.write('gas_specs_names\t'.expandtabs(26) + " ".join(str(x) for x in gas_specs_names) + '\n')

                # Write gas energies and molecular weights
                tags_dict = ['gas_energy', 'gas_molec_weight']
                tags_zacros = ['gas_energies', 'gas_molec_weights']
                for tag1, tag2 in zip(tags_dict, tags_zacros):
                    tag_list = [self.gas_model.df.loc[x, tag1] for x in gas_specs_names]
                    if tag1 == 'gas_energy':
                        formatted_tag_list = [f'{x:.{sig_figs_energies}f}' for x in tag_list]
                        infile.write(f'{tag2}\t'.expandtabs(26) + " ".join(formatted_tag_list) + '\n')
                    else:
                        infile.write(f'{tag2}\t'.expandtabs(26) + " ".join(str(x) for x in tag_list) + '\n')

                # Write gas molar fractions
                try:
                    gas_molar_frac_list = [pressure[x] / p_tot for x in gas_specs_names]
                except KeyError as ke:
                    print(f"Key not found in 'pressure' dictionary: {ke}")
                    print(f"When calling KMCModel.create_job_dir(), 'pressure' dictionary must contain the names of all "
                          f"gas species ")
                    gas_molar_frac_list = [0.0 for _ in gas_specs_names]  # Assign zero fractions for missing species

                infile.write(f'gas_molar_fracs\t'.expandtabs(26) + " ".join(str(x) for x in gas_molar_frac_list) + '\n')

                # Write number of surface species and their names and dentates
                infile.write('n_surf_species\t'.expandtabs(26) + str(len(surf_specs)) + '\n')
                infile.write('surf_specs_names\t'.expandtabs(26) + " ".join(str(x) for x in surf_specs.keys()) + '\n')
                infile.write('surf_specs_dent\t'.expandtabs(26) + " ".join(str(x) for x in surf_specs.values()) + '\n')

                # Write reporting scheme
                for tag in ['snapshots', 'process_statistics', 'species_numbers']:
                    infile.write((tag + '\t').expandtabs(26) + str(reporting_scheme.get(tag, '')) + '\n')

                # Write stopping criteria
                for tag in ['max_steps', 'max_time', 'wall_time']:
                    infile.write((tag + '\t').expandtabs(26) + str(stopping_criteria.get(tag, '')) + '\n')

                # Handle stiffness scaling based on version:
                if float(version) >= 5.0:
                    if stiffness_scalable_steps or stiffness_scalable_symmetric_steps:
                        if stiffness_scaling_algorithm is None:
                            infile.write("enable_stiffness_scaling\n")
                        else:
                            infile.write('enable_stiffness_scaling\t'.expandtabs(26) + stiffness_scaling_algorithm + '\n')
                        for tag in stiffness_scaling_tags:
                            infile.write((tag + '\t').expandtabs(26) + str(stiffness_scaling_tags[tag]) + '\n')
                else:
                    if stiffness_scalable_steps:
                        infile.write("enable_stiffness_scaling\n")
                        for tag in stiffness_scaling_tags:
                            infile.write((tag + '\t').expandtabs(26) + str(stiffness_scaling_tags[tag]) + '\n')

                # Write any additional keywords
                for kw in additional_keywords:
                    infile.write(f"{kw}\n")

                infile.write(f"finish\n")
        except IOError as e:
            raise KMCModelError(f"Failed to write to 'simulation_input.dat': {e}")

    def get_surf_specs(self):
        """
        Identify all surface species and their corresponding dentates from the `energetics_model` DataFrame.

        Used to write `'surf_specs_names'` and `'surf_specs_dent'` in the `simulation_input.dat` file.

        Returns
        -------
        dict
            A dictionary with surface species names as keys and dentates as values.

        Raises
        ------
        KMCModelError
            If the `lattice_state` format is invalid.
        """
        surf_specs = {}
        for cluster in self.energetics_model.df.index:
            lattice_state = self.energetics_model.df.loc[cluster, 'lattice_state']
            for site in lattice_state:
                # Assuming the format is '1 CO* 1' or similar
                parts = site.split()
                if len(parts) >= 3:
                    surf_specs_name = parts[1]
                    try:
                        surf_specs_dent = int(parts[2])
                    except ValueError:
                        raise KMCModelError(
                            f"Invalid dentate value in lattice_state for cluster '{cluster}': {parts[2]}")
                    if surf_specs_name not in surf_specs or (
                            surf_specs_name in surf_specs and surf_specs_dent > surf_specs[surf_specs_name]):
                        surf_specs[surf_specs_name] = surf_specs_dent
                else:
                    raise KMCModelError(
                        f"Invalid lattice_state format for cluster '{cluster}': {site}")
        return surf_specs



# tests/test_calc_functions.py

import math
import numpy as np
import pytest

from zacrostools import calc_functions as cf


def test_as_eV_list():
    assert cf.as_eV_list([1000, 0, 250]) == [1.0, 0.0, 0.25]
    assert cf.as_eV_list([]) == []
    assert cf.as_eV_list(None) == []


def test_q_vib_numeric():
    # T = 300 K, vib energies in meV
    T = 300.0
    vib = [100, 200, 300]
    # Known value (from the HO formula used here)
    expected = 9.32366938463989e-06
    got = cf.q_vib(vib, T)
    assert np.isclose(got, expected, rtol=1e-12), f"q_vib mismatch: {got} vs {expected}"


def test_q_trans2D_numeric():
    # A = 5 Å^2, m = 28 amu, T = 300 K
    T = 300.0
    A = 5.0
    m_amu = 28.0
    expected = 137.80008332596037  # dimensionless
    got = cf.q_trans2D(A, m_amu, T)
    assert np.isclose(got, expected, rtol=1e-12), f"q_trans2D mismatch: {got} vs {expected}"


def test_q_rot_linear_numeric():
    # Linear rotor: 1 inertia moment
    T = 300.0
    inertia = [10.0]  # amu*Å^2
    sigma = 2
    expected = 61.84453277727122
    got = cf.q_rot(inertia, sigma, T)
    assert np.isclose(got, expected, rtol=1e-12), f"q_rot linear mismatch: {got} vs {expected}"


def test_q_rot_nonlinear_numeric():
    # Non-linear rotor: 3 inertia moments
    T = 300.0
    inertia = [10.0, 15.0, 20.0]  # amu*Å^2
    sigma = 1
    expected = 4223.111291268305
    got = cf.q_rot(inertia, sigma, T)
    assert np.isclose(got, expected, rtol=1e-12), f"q_rot nonlinear mismatch: {got} vs {expected}"


def test_q_rot_invalid_raises():
    T = 300.0
    bad_inertia = [10.0, 15.0]  # invalid length
    with pytest.raises(ValueError):
        cf.q_rot(bad_inertia, sym_number=1, T=T)


def test_q_elec_defaults_and_values():
    assert cf.q_elec(None) == 1.0
    assert cf.q_elec(0) == 1.0
    assert cf.q_elec(1) == 1.0
    assert cf.q_elec(3) == 3.0


def test_gas_RS_partition_composition():
    # Check that gas_RS_partition equals product of components
    T = 300.0
    A = 5.0
    molec = {
        "gas_molec_weight": 28.0,      # amu
        "inertia_moments": [10.0],     # linear
        "sym_number": 2,
        "degeneracy": 1,
    }
    vib_RS = [100, 200, 300]

    # Build product explicitly
    q_trans = cf.q_trans2D(A, molec["gas_molec_weight"], T)
    q_rot = cf.q_rot(molec["inertia_moments"], molec["sym_number"], T)
    q_vib = cf.q_vib(vib_RS, T)
    q_elec = cf.q_elec(molec["degeneracy"])
    expected = q_trans * q_rot * q_vib * q_elec

    got = cf.gas_RS_partition(A, molec, vib_RS, T)
    assert np.isclose(got, expected, rtol=1e-12), "gas_RS_partition does not match product of components"


def test_pe_surface_numeric():
    # Surface process: pe = (kB*T/h) * (q_TS/q_RS)
    T = 300.0
    vib_RS = [100, 200, 300]
    vib_TS = [150, 250, 350]
    expected = 337272372355.2879  # s^-1
    got = cf.pe_surface(vib_TS_meV=vib_TS, vib_RS_meV=vib_RS, T=T)
    assert np.isclose(got, expected, rtol=1e-10), f"pe_surface mismatch: {got} vs {expected}"


def test_pe_activated_ads_numeric():
    # pe = A / sqrt(2*pi*m*k*T) * (q_TS / q_RS)
    T = 300.0
    A = 5.0
    molec = {
        "gas_molec_weight": 28.0,  # amu
        "inertia_moments": [10.0],
        "sym_number": 2,
        "degeneracy": 1,
    }
    vib_RS = [100, 200, 300]
    vib_TS = [120, 220, 320]
    expected = 5.22368430437502e-02  # s^-1 Pa^-1
    got = cf.pe_activated_ads(A_ang2=A, molec_data=molec, vib_TS_meV=vib_TS, vib_RS_meV=vib_RS, T=T)
    assert np.isclose(got, expected, rtol=1e-12), f"pe_activated_ads mismatch: {got} vs {expected}"


def test_pe_nonactivated_ads_numeric():
    # pe = A / sqrt(2*pi*m*k*T)
    T = 300.0
    A = 5.0
    mass_amu = 28.0
    expected = 1437.3887352217173  # s^-1 Pa^-1
    got = cf.pe_nonactivated_ads(A_ang2=A, mass_amu=mass_amu, T=T)
    assert np.isclose(got, expected, rtol=1e-12), f"pe_nonactivated_ads mismatch: {got} vs {expected}"


def test_pe_nonactivated_desorption_numeric():
    # pe = (kB*T/h) * (q_PS / q_RS)
    # q_PS = q_trans2D * q_rot * q_vib * q_elec (gas-like, uses PS data)
    # q_RS = q_vib (surface)
    T = 300.0
    A = 5.0
    molec_ps = {
        "gas_molec_weight": 28.0,  # amu
        "inertia_moments": [10.0],  # linear
        "sym_number": 2,
        "degeneracy": 1,
    }
    vib_PS = [150, 250, 350]
    vib_RS_surface = [100, 200, 300]
    expected = 2.874296463198428e15  # s^-1
    got = cf.pe_nonactivated_desorption(
        A_ang2=A,
        molec_data=molec_ps,
        vib_PS_meV=vib_PS,
        vib_RS_meV_surface=vib_RS_surface,
        T=T
    )
    assert np.isclose(got, expected, rtol=1e-12), f"pe_nonactivated_desorption mismatch: {got} vs {expected}"



import os
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm, SymLogNorm, Normalize
from matplotlib.ticker import FuncFormatter
import matplotlib.patheffects as pe
import matplotlib.colors as colors

from zacrostools.kmc_output import KMCOutput
from zacrostools.custom_exceptions import PlotError
from zacrostools.heatmaps.heatmap_functions import get_axis_label, extract_value, convert_to_subscript
from zacrostools.detect_issues import detect_issues


def plot_dtof(
        # general mandatory parameters
        ax,
        x: str,
        y: str,
        scan_path: str,
        # plot-specific mandatory parameters
        gas_spec: str = None,
        scan_path_ref: str = None,
        # plot-specific optional parameters
        difference_type: str = 'absolute',
        check_issues: str = 'none',
        scale: str = 'log',
        min_molec: int = 1,
        max_dtof: float = None,
        min_dtof: float = None,
        min_tof_ref: float = 0.0,
        nlevels: int = 0,
        weights: str = None,
        analysis_range: list = None,
        range_type: str = 'time',
        # detect_issues optional parameters (pass-through)
        energy_slope_thr: float = 5.0e-10,
        time_r2_thr: float = 0.95,
        max_points: int = 100,
        # general optional parameters
        cmap: str = "RdYlBu",
        show_points: bool = False,
        show_colorbar: bool = True,
        auto_title: bool = False, **kwargs):
    """
    Plot the change in Turnover Frequency (∆TOF) heatmap between a main simulation and a reference.

    This function reads KMC simulation outputs for a specified gas species from both a main set of
    directories (`scan_path`) and a reference set (`scan_path_ref`), computes the TOF for each, and
    then builds a 2D heatmap of the difference (∆TOF). It supports absolute differences or relative
    differences, and can optionally mask out any simulations flagged as having issues.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Matplotlib axis to draw the heatmap on.
    x, y : str
        Names of the parameters defining the horizontal and vertical axes (e.g., 'pressure_CH4',
        'pressure_CO2'). If the string contains 'pressure', the respective axis will be logarithmic.
    scan_path : str
        Path to the folder containing the main simulation subdirectories.
    gas_spec : str, optional
        Gas species key (e.g., 'H2', 'CO') for which TOF is computed.
    scan_path_ref : str, optional
        Path to the folder containing reference simulation subdirectories (must match names).
    difference_type : {'absolute', 'relative'}, default 'absolute'
        - 'absolute': ∆TOF = TOF(main) - TOF(ref)
        - 'relative': ∆TOF = | TOF(main) / TOF(ref) |
          In this mode, the colorbar is always logarithmic and centered at 10^0 (i.e., ratio 1).
    check_issues : {'none','both','main','ref'}, default 'none'
        Which runs `detect_issues` on:
        - 'none' → neither
        - 'both' → both main and ref
        - 'main' → only the main scan
        - 'ref'  → only the reference scan
    scale : {'log', 'lin'}, default 'log'
        Axis scaling for the heatmap. (Forced to 'log' if difference_type='relative'.)
    min_molec : int, default 1
        Minimum total production threshold for both main and reference runs; below this, cell
        is masked.
    max_dtof, min_dtof : float, optional
        Maximum and minimum threshold for the color scale. If None, sensible defaults are chosen:
        - For absolute differences: max → next decade above data, min → max/1e3.
        - For relative (ratio): the colorbar spans [10^{-N}, 10^{+N}] around 1, where 10^N is the
          nearest order of magnitude to `max_dtof` if provided, otherwise to the maximum ratio in data.
    min_tof_ref : float, optional
        Minimum TOF threshold for the reference run. If nonzero and the reference TOF is below
        this value, the corresponding cell is masked (NaN). Default is 0.0.
    nlevels : int, default 0
        Number of discrete color levels (must be odd ≥3 for absolute; any int ≥2 for relative). If 0,
        continuous normalization is used.
    weights : str, optional
        Weighting mode passed to KMCOutput ('time', 'events', or None).
    analysis_range : list of two floats, default [0,100]
        Percentage of simulation range to analyze (start, end).
    range_type : {'time', 'nevents'}, default 'time'
        Whether analysis_range refers to simulation time or number of events.
    energy_slope_thr : float, optional
        Threshold for the absolute energy slope used by `detect_issues` (default is 5.0e-10).
    time_r2_thr : float, optional
        R² threshold for the linearity of time vs. events used by `detect_issues` (default is 0.95).
    max_points : int, optional
        Maximum number of data points sampled by `detect_issues` (default is 100).
    cmap : str, default 'RdYlBu'
        Colormap for the heatmap.
    show_points : bool, default False
        If True, overlays scatter points at each grid location.
    show_colorbar : bool, default True
        Whether to draw a colorbar.
    auto_title : bool, default False
        If True, automatically set the plot title indicating ∆TOF and species.

    Notes
    -----
    - The function reads simulation data from the provided main and reference directories. It computes the TOF
      for a given gas species in both sets of simulations and calculates ∆TOF.
    """
    # Parameter removal notice
    if 'percent' in kwargs:
        raise ValueError("'percent' is not supported for dtof since v2.10; relative ∆TOF is now defined as |TOF(main)/TOF(ref)| (unitless ratio).")

    # Set default analysis range if needed
    if analysis_range is None:
        analysis_range = [0, 100]

    if check_issues not in ('none', 'both', 'main', 'ref'):
        raise ValueError(f"Incorrect value for check_issues: {check_issues}")

    # Validate scan_path and scan_path_ref
    if not os.path.isdir(scan_path):
        raise ValueError(f"Scan path folder does not exist: {scan_path}")
    if not os.path.isdir(scan_path_ref):
        raise ValueError(f"Reference scan path folder does not exist: {scan_path_ref}")

    simulation_dirs = glob.glob(os.path.join(scan_path, "*"))
    if len(simulation_dirs) == 0:
        raise ValueError(f"Scan path folder is empty: {scan_path}")

    simulation_ref_dirs = glob.glob(os.path.join(scan_path_ref, "*"))
    if len(simulation_ref_dirs) == 0:
        raise ValueError(f"Scan path folder is empty: {scan_path_ref}")

    # Determine whether x and y values are logarithmic (based on presence of 'pressure' in the variable name)
    x_is_log = "pressure" in x
    y_is_log = "pressure" in y

    # Initialize lists and DataFrame to store data
    x_value_list, y_value_list = [], []
    df = pd.DataFrame()

    # Loop over simulation directories (using matching folder names in the reference directory)
    for sim_path in simulation_dirs:
        folder_name = os.path.basename(sim_path)
        ref_path = os.path.join(scan_path_ref, folder_name)

        # Extract x and y values
        x_value = extract_value(x, sim_path)
        y_value = extract_value(y, sim_path)
        df.loc[folder_name, "x_value"] = x_value
        df.loc[folder_name, "y_value"] = y_value
        if x_value not in x_value_list:
            x_value_list.append(x_value)
        if y_value not in y_value_list:
            y_value_list.append(y_value)

        # Compute TOF and ∆TOF
        try:
            kmc = KMCOutput(
                path=sim_path,
                analysis_range=analysis_range,
                range_type=range_type,
                weights=weights,
            )
            kmc_ref = KMCOutput(
                path=ref_path,
                analysis_range=analysis_range,
                range_type=range_type,
                weights=weights,
            )

            tof = max(kmc.tof[gas_spec], 0.0)
            prod = kmc.total_production[gas_spec]
            tof_ref = max(kmc_ref.tof[gas_spec], 0.0)
            prod_ref = kmc_ref.total_production[gas_spec]

            # Store production for later use (even if we end up masking the point)
            df.loc[folder_name, "total_production"] = prod
            df.loc[folder_name, "total_production_ref"] = prod_ref

            has_main_issues = False
            has_ref_issues = False

            if check_issues in ['both', 'main']:
                has_main_issues = detect_issues(
                    job_path=sim_path,
                    analysis_range=analysis_range,
                    range_type=range_type,
                    energy_slope_thr=energy_slope_thr,
                    time_r2_thr=time_r2_thr,
                    max_points=max_points)

            if check_issues in ['both', 'ref']:
                has_ref_issues = detect_issues(
                    job_path=ref_path,
                    analysis_range=analysis_range,
                    range_type=range_type,
                    energy_slope_thr=energy_slope_thr,
                    time_r2_thr=time_r2_thr,
                    max_points=max_points)

            if check_issues == 'both':
                if has_main_issues or has_ref_issues:
                    df.loc[folder_name, "dtof"] = np.nan
                    continue
            elif check_issues == 'main':
                if has_main_issues:
                    df.loc[folder_name, "dtof"] = np.nan
                    continue
            elif check_issues == 'ref':
                if has_ref_issues:
                    df.loc[folder_name, "dtof"] = np.nan
                    continue

            # If a minimum reference TOF is requested, mask cells where the reference is below it
            if (min_tof_ref != 0.0) and (tof_ref < min_tof_ref):
                df.loc[folder_name, "dtof"] = np.nan
                continue

            # Only consider points above production threshold
            if prod >= min_molec and prod_ref >= min_molec:
                if difference_type == "absolute":
                    dtof_val = tof - tof_ref
                elif difference_type == "relative":
                    dtof_val = (abs(tof / tof_ref) if tof_ref != 0 else np.nan)
                else:
                    raise ValueError(
                        "difference_type must be 'absolute' or 'relative'"
                    )
            else:
                dtof_val = np.nan

            df.loc[folder_name, "dtof"] = dtof_val

        except Exception as e:
            print(f"Warning: could not process {folder_name}: {e}")
            df.loc[folder_name, ["dtof",
                                 "total_production",
                                 "total_production_ref"]] = np.nan

    # Build sorted arrays for x and y axis values
    x_value_list = np.sort(np.asarray(x_value_list))
    y_value_list = np.sort(np.asarray(y_value_list))
    x_list = np.power(10, x_value_list) if x_is_log else x_value_list
    y_list = np.power(10, y_value_list) if y_is_log else y_value_list

    # Create a 2D grid
    z_axis = np.full((len(y_value_list), len(x_value_list)), np.nan)

    for i, x_val in enumerate(x_value_list):
        for j, y_val in enumerate(y_value_list):
            matches = df[(df["x_value"] == x_val) & (df["y_value"] == y_val)].index
            if len(matches) > 1:
                raise PlotError(
                    f"Several folders share {x} = {x_val} and {y} = {y_val}"
                )
            elif len(matches) == 0:
                print(
                    f"Warning: folder for x = {x_val} and y = {y_val} missing; NaN assigned"
                )
                continue

            folder = matches[0]
            dtof_val = df.loc[folder, "dtof"]

            if np.isnan(dtof_val):
                continue  # leave as NaN

            # Optional value capping (absolute vs relative handled later when N is known)
            if (difference_type == "absolute") and (max_dtof is not None):
                if dtof_val > max_dtof:
                    dtof_val = max_dtof
                elif dtof_val < -max_dtof:
                    dtof_val = -max_dtof

            z_axis[j, i] = dtof_val

    x_axis, y_axis = np.meshgrid(x_list, y_list)

    # --- Determine normalization parameters ---
    levels = None

    if difference_type == 'absolute':
        # Compute a symmetric maximum value from the grid.
        computed_abs_max = max(np.abs(np.nanmin(z_axis)), np.abs(np.nanmax(z_axis)))
        if max_dtof is None:
            exponent = np.ceil(np.log10(computed_abs_max)) if np.isfinite(computed_abs_max) and computed_abs_max > 0 else 0
            max_dtof = 10 ** exponent
        if min_dtof is None:
            min_dtof = max_dtof / 1.0e3
        abs_max = max_dtof

        # Normalization (continuous/discrete)
        if nlevels == 0:
            if scale == "lin":
                norm = Normalize(vmin=-abs_max, vmax=abs_max)
            elif scale == "log":
                if np.all(z_axis > 0):
                    norm = LogNorm(vmin=min_dtof, vmax=abs_max)
                elif np.all(z_axis < 0):
                    norm = LogNorm(vmin=-abs_max, vmax=-min_dtof)
                else:
                    norm = SymLogNorm(
                        linthresh=min_dtof,
                        linscale=1.0,
                        vmin=-abs_max,
                        vmax=abs_max,
                        base=10,
                    )
            else:
                raise ValueError("scale must be 'log' or 'lin'")
        else:
            if (not isinstance(nlevels, int)) or (nlevels < 3) or (nlevels % 2 == 0):
                raise ValueError("nlevels must be either 0 or a positive odd integer (≥3)")
            if scale == 'lin':
                levels = np.linspace(-abs_max, abs_max, nlevels)
            elif scale == 'log':
                positive_levels = np.logspace(np.log10(min_dtof), np.log10(abs_max), (nlevels - 1) // 2)
                levels = np.concatenate((-positive_levels[::-1], [0], positive_levels))
            else:
                raise ValueError("scale parameter must be either 'log' or 'lin'")
            norm = colors.BoundaryNorm(levels, ncolors=plt.get_cmap(cmap).N, clip=True)

    else:
        # RELATIVE (ratio): force log scale centered at 10^0
        scale = 'log'

        # Determine data max (ratio) and choose nearest order of magnitude
        data_max = np.nanmax(z_axis)
        if not np.isfinite(data_max) or data_max <= 0:
            data_max = 1.0

        if max_dtof is None:
            N = int(np.round(np.log10(data_max)))
        else:
            N = int(np.round(np.log10(max_dtof)))
        # Use symmetric log-range [10^{-N}, 10^{+N}]
        vmin = 10.0 ** (-abs(N))
        vmax = 10.0 ** (abs(N))

        # After N is known, cap values into [10^{-N}, 10^{N}]
        for j in range(z_axis.shape[0]):
            for i in range(z_axis.shape[1]):
                val = z_axis[j, i]
                if np.isnan(val):
                    continue
                if val < vmin:
                    z_axis[j, i] = vmin
                elif val > vmax:
                    z_axis[j, i] = vmax

        if nlevels == 0:
            norm = LogNorm(vmin=vmin, vmax=vmax)
        else:
            # For ratios we don't require odd nlevels; any >=2 works. Keep behavior minimal.
            if (not isinstance(nlevels, int)) or (nlevels < 2):
                raise ValueError("For relative mode, nlevels must be either 0 or an integer ≥ 2")
            levels = np.logspace(-abs(N), abs(N), nlevels)
            norm = colors.BoundaryNorm(levels, ncolors=plt.get_cmap(cmap).N, clip=True)

    # --- Create pcolormesh plot ---
    cp = ax.pcolormesh(x_axis, y_axis, z_axis, cmap=cmap, norm=norm)

    if show_colorbar:
        if levels is not None:
            cbar = plt.colorbar(cp, ax=ax, boundaries=levels, ticks=levels)
        else:
            cbar = plt.colorbar(cp, ax=ax)
        # --- Custom formatter for colorbar tick labels ---
        if scale == 'log':
            if difference_type == 'relative':
                def rel_log_formatter(x, pos):
                    # Show 10^{k}, centered at 10^0
                    if x <= 0 or not np.isfinite(x):
                        return ''
                    k = int(np.round(np.log10(x)))
                    # Only show pretty powers of ten; otherwise scientific
                    if np.isclose(x, 10 ** k, rtol=1e-5, atol=1e-12):
                        return r'$10^{%d}$' % k
                    return r'$%1.1e$' % x
                formatter = FuncFormatter(rel_log_formatter)
            else:
                def log_formatter(x, pos):
                    if np.isclose(x, 0):
                        return '0'
                    exponent_cb = int(np.floor(np.log10(abs(x))))
                    if np.isclose(abs(x), 10 ** exponent_cb, rtol=1e-5, atol=1e-8):
                        return r'$+10^{%d}$' % exponent_cb if x > 0 else r'$-10^{%d}$' % exponent_cb
                    else:
                        return r'$%+1.1e$' % x
                formatter = FuncFormatter(log_formatter)
        else:
            formatter = FuncFormatter(
                lambda x, pos: "0" if np.isclose(x, 0)
                else f"{x:+.0f}"
            )
        cbar.ax.yaxis.set_major_formatter(formatter)
        # ---------------------------------------------------

    ax.set_xlim(np.min(x_list), np.max(x_list))
    ax.set_ylim(np.min(y_list), np.max(y_list))
    ax.set_xscale('log' if x_is_log else 'linear')
    ax.set_yscale('log' if y_is_log else 'linear')
    ax.set_xlabel(get_axis_label(x))
    ax.set_ylabel(get_axis_label(y))
    ax.set_facecolor("lightgray")

    if auto_title:
        _tag = "abs" if difference_type == "absolute" else "rel"
        label = f"$\\Delta\\mathrm{{TOF}}_{{{_tag}}}\\ {convert_to_subscript(gas_spec)}$"
        ax.set_title(
            label=label,
            y=1.0,
            pad=-14,
            color="w",
            path_effects=[pe.withStroke(linewidth=2, foreground="black")]
        )

    if show_points:
        ax.plot(x_axis.flatten(), y_axis.flatten(), 'w.', markersize=3)

    return cp



import functools
import inspect
from typing import Union, get_origin, get_args


class ZacrosToolsError(Exception):
    """Base class for exceptions in this library."""
    pass


class LatticeModelError(ZacrosToolsError):
    """Exception raised for errors in the lattice model."""
    pass


class GasModelError(ZacrosToolsError):
    """Exception raised for errors in the gas model."""
    pass


class EnergeticsModelError(ZacrosToolsError):
    """Exception raised for errors in the energetics model."""
    pass


class ReactionModelError(ZacrosToolsError):
    """Exception raised for errors in the reaction model."""
    pass


class KMCModelError(ZacrosToolsError):
    """Exception raised for errors when creating the KMC model."""
    pass


class CalcFunctionsError(ZacrosToolsError):
    """Exception raised for errors in the calc_functions module."""
    pass


class KMCOutputError(ZacrosToolsError):
    """Exception raised for errors in the kmc_output module."""
    pass


class PlotError(ZacrosToolsError):
    """Exception raised for errors in the plot_functions module."""
    pass


def enforce_types(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        signature = inspect.signature(func)
        bound_args = signature.bind(*args, **kwargs)
        for name, value in bound_args.arguments.items():
            expected_type = signature.parameters[name].annotation
            if expected_type != inspect.Parameter.empty:
                if get_origin(expected_type) is Union:
                    allowed_types = get_args(expected_type)
                else:
                    allowed_types = (expected_type,)
                if not any(isinstance(value, t) for t in allowed_types):
                    expected_types_str = ', '.join([t.__name__ for t in allowed_types])
                    raise TypeError(
                        f"Argument '{name}' must be one of types: {expected_types_str}; "
                        f"got {type(value).__name__} instead."
                    )
        return func(*args, **kwargs)

    return wrapper



import os
import glob
from typing import Union, Optional
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patheffects as pe
from matplotlib.colors import LogNorm

from zacrostools.kmc_output import KMCOutput
from zacrostools.custom_exceptions import PlotError
from zacrostools.heatmaps.heatmap_functions import get_axis_label, extract_value, convert_to_subscript


def plot_tof(
        # general mandatory parameters
        ax,
        x: str,
        y: str,
        scan_path: str,
        # plot-specific mandatory parameters
        gas_spec: str,
        # plot-specific optional parameters
        min_molec: int = 1,
        weights: str = None,
        levels: Optional[Union[list, np.ndarray]] = None,
        analysis_range: list = None,
        range_type: str = 'time',
        show_max: bool = False,
        # general optional parameters
        cmap: str = "inferno",
        show_points: bool = False,
        show_colorbar: bool = True,
        auto_title: bool = False):
    """
    Plot a Turnover Frequency (TOF) heatmap using a contourf plot.

    Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axis object where the plot is drawn.
        x : str
            Parameter name for the x-axis (e.g., 'pressure_X' or 'temperature').
        y : str
            Parameter name for the y-axis.
        scan_path : str
            Path to the directory containing simulation subdirectories.
        gas_spec : str
            Gas species for which the TOF is calculated.
        min_molec : int, optional
            Minimum number of molecules required for a valid TOF value.
        weights : str, optional
            Weighting method for TOF calculation (e.g., 'time', 'events', or None).
        levels : list or np.ndarray, optional
            Contour levels to use in the plot. When provided, TOF values are clipped to the
            minimum and maximum values in this list.
        analysis_range : list, optional
            Portion of the simulation data to analyze (default is [0, 100]).
        range_type : str, optional
            Type of range to consider in the analysis ('time' or 'nevents').
        cmap : str, optional
            Colormap to be used for the contour plot (default is 'inferno').
        show_max : bool, optional
            If True, mark the simulation point with the highest TOF using a golden star.
        show_points : bool, optional
            If True, overlay the grid points on the heatmap.
        show_colorbar : bool, optional
            If True, display a colorbar alongside the heatmap.
        auto_title : bool, optional
            If True, automatically set the plot title based on the gas species.

    Notes
    -----
    - TOF values are assigned as NaN if the corresponding simulation data is missing or if
      the total production does not meet the minimum molecule threshold.
    """

    # Set default analysis range if needed
    if analysis_range is None:
        analysis_range = [0, 100]

    # Validate scan_path
    if not os.path.isdir(scan_path):
        raise ValueError(f"Scan path folder does not exist: {scan_path}")

    simulation_dirs = glob.glob(os.path.join(scan_path, "*"))
    if len(simulation_dirs) == 0:
        raise ValueError(f"Scan path folder is empty: {scan_path}")

    # Determine whether x and y values are logarithmic (based on presence of 'pressure' in the variable name)
    x_is_log = "pressure" in x
    y_is_log = "pressure" in y

    # Initialize lists and DataFrame to store data
    x_value_list, y_value_list = [], []
    df = pd.DataFrame()

    # Collect simulation data in a dataframe
    for sim_path in simulation_dirs:
        folder_name = os.path.basename(sim_path)

        # Extract x and y values
        x_value = extract_value(x, sim_path)
        y_value = extract_value(y, sim_path)
        df.loc[folder_name, "x_value"] = x_value
        df.loc[folder_name, "y_value"] = y_value
        if x_value not in x_value_list:
            x_value_list.append(x_value)
        if y_value not in y_value_list:
            y_value_list.append(y_value)

        # Initialize KMCOutput and retrieve TOF and total production for the given gas_spec
        try:
            kmc_output = KMCOutput(
                path=sim_path,
                analysis_range=analysis_range,
                range_type=range_type,
                weights=weights)
            df.loc[folder_name, "tof"] = max(kmc_output.tof[gas_spec], 0.0)  # only consider positive TOF
            df.loc[folder_name, "total_production"] = kmc_output.total_production[gas_spec]
        except Exception as e:
            print(f"Warning: Could not initialize KMCOutput for {folder_name}: {e}")
            df.loc[folder_name, "tof"] = float('NaN')

    # Build sorted arrays for x and y axis values
    x_value_list = np.sort(np.asarray(x_value_list))
    y_value_list = np.sort(np.asarray(y_value_list))
    x_list = np.power(10, x_value_list) if x_is_log else x_value_list
    y_list = np.power(10, y_value_list) if y_is_log else y_value_list

    # Create a 2D grid
    z_axis = np.full((len(y_value_list), len(x_value_list)), np.nan)
    for i, x_val in enumerate(x_value_list):
        for j, y_val in enumerate(y_value_list):
            matching_indices = df[(df['x_value'] == x_val) & (df['y_value'] == y_val)].index
            if len(matching_indices) > 1:
                raise PlotError(
                    f"Several folders have the same values of {x} ({x_val}) and {y} ({y_val})")
            elif len(matching_indices) == 0:
                print(f"Warning: folder for x = {x_val} and y = {y_val} missing, NaN assigned")
            else:
                folder_name = matching_indices[0]

                # Only assign the TOF if total production meets the threshold
                if df.loc[folder_name, "total_production"] >= min_molec:

                    # Apply a minimum and maximum value if levels are provided
                    if levels is not None:
                        if df.loc[folder_name, "tof"] < min(levels):
                            z_axis[j, i] = min(levels)
                        elif df.loc[folder_name, "tof"] > max(levels):
                            z_axis[j, i] = max(levels)
                        else:
                            z_axis[j, i] = df.loc[folder_name, "tof"]
                    else:
                        z_axis[j, i] = df.loc[folder_name, "tof"]

    x_axis, y_axis = np.meshgrid(x_list, y_list)

    # Create a contourf plot using logarithmic normalization
    if levels is not None:
        cp = ax.contourf(x_axis, y_axis, z_axis, levels=levels, cmap=cmap,
                         norm=LogNorm(vmin=min(levels), vmax=max(levels)))
    else:
        cp = ax.contourf(x_axis, y_axis, z_axis, cmap=cmap, norm=LogNorm())

    # Optionally, mark the maximum TOF point with a golden star
    if show_max:
        max_index = np.nanargmax(z_axis)
        max_j, max_i = np.unravel_index(max_index, z_axis.shape)
        max_x = x_axis[max_j, max_i]
        max_y = y_axis[max_j, max_i]
        ax.plot(max_x, max_y, marker='*', color='gold', markersize=8,
                path_effects=[pe.withStroke(linewidth=2, foreground="black")])

    if show_colorbar:
        cbar = plt.colorbar(cp, ax=ax)

    ax.set_xlim(np.min(x_list), np.max(x_list))
    ax.set_ylim(np.min(y_list), np.max(y_list))
    ax.set_xscale('log' if x_is_log else 'linear')
    ax.set_yscale('log' if y_is_log else 'linear')
    ax.set_xlabel(get_axis_label(x))
    ax.set_ylabel(get_axis_label(y))
    ax.set_facecolor("lightgray")

    if auto_title:
        ax.set_title(
            label="TOF " + f"${convert_to_subscript(chemical_formula=gas_spec)}$",
            y=1.0,
            pad=-14,
            color="w",
            path_effects=[pe.withStroke(linewidth=2, foreground="black")]
        )

    if show_points:
        ax.plot(x_axis.flatten(), y_axis.flatten(), 'w.', markersize=3)

    return cp
